/*    ILFIVE lisp objects library -- matching s-espressions with syntax specs
 *
 *    Copyright (c) 1991--1998 Alexandru Dan Corlan M.D.
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    In short, you can use and modify this program. You can
 *    distribute the program with or without your modifications if you want,
 *    but you must not require licence fees and must provide the
 *    full source code with the distribution. If you make an application program
 *    which just calls functions from this library you may distribute that
 *    application program without source code, but also as an linkable object file, 
 *    and must distribute also this library with source code.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* PRIMITIVES DEFINED IN THIS PACKAGE (which requires linking of 
 * the il5ev package for primitive definitions; il5ev() must be 
 * called before il5check()).
 *
 * (check <object> <expression>)
 *   -> returns either a binary if check succeeded, either a string
 *      representing an error message or the 
 *      syntax_error_ = :syntax-error symbol representing the fact
 *      that there was an error having no associated message
 *      or meta_syntax_error_ meaning that some pattern is 
 *      built wrong (but not all malconstructions are detected,
 *      only some 5% of them)
 *
 * where <expression> is a Metapat V2.0 expression:
 *
    :nothing                       --- just ignored (to be used in 
                                       :and)
    :anything                      --- one object, no matter which
    :any-symbol
    :any-binary
    :any-string
    :any-list
    :any-vector
    :any-float
 	 ( <expr> <expr> ... )          --- a list of expressions
    #( <expr> <expr> ... )         --- a vector of expressions
    :the-symbol <symbol>           --- that very item
    :or ( <expr> <expr> ... )      --- one of the expressions (the
                                       first that is successfully 
                                       matched decides how many items
                                       are lifted from the stream, if
                                       it is to speak of a :sequence 
                                       --- see below)
    :and ( <expr> <expr> ...)      --- all must match; the error
                                       symbol or message returned
                                       by the first not matching one
                                       is returned; the smallest
                                       number of them are passed over
    :zero-or-more <expr>           --- none, one or more that one
    :one-or-more <expr>				  --- one or more that one
    :not <expr>						  --- the following must not be 
                                       validated by expr; as many items
                                       as expr tells are passed over
    :set-of (<expr> <expr> ...)	  --- zero or more occourences of
                                       the expressions, in any order,
                                       but each one occouring at most 
                                       once; if none is present
                                       it passes to the next item
    :sequence (<expr> <expr> ...)  --- the expressions must come in
                                       sequence (usefull especially
                                       in conjuction with :one-of
                                       and :set-of)
    :check <fun> <expr>            --- the function is called with
                                       the following item and the given
                                       expression
    :otherwise "format string"     --- can not follow after a :check clause
 *
 */

#include <stdio.h>
#include <malloc.h>

#include <shpell5.h>
cell strbuild(char *s);
extern cell no_sym_val_;

cell anything_;
cell nothing_;
cell any_symbol_;
cell any_binary_;
cell any_string_;
cell any_list_;
cell any_vector_;
cell any_float_;
cell any_keyword_;
cell the_symbol_;
cell or_;
cell and_;
cell zero_more_;
cell one_more_;
cell not_;
cell set_of_;
cell sequence_;
cell check_;
cell rest_check_;
cell othw_;
cell othw_append_;
cell othw_replace_;

cell optional_;
cell required_;

cell syntax_error_;
cell meta_syntax_error_;

static cell nr[10];  
    /* With numbers to be used as return values */
    /* in an otherwise optimal way */

static char il5cbuffer[1024];


cell lappend(cell l1, cell l2)

  {
  if(l1==NIL)
    return l2;
  return cons(car(l1), lappend(cdr(l1), l2));
  }

/* advancing a list pattern */

#define is_otherwise(c) (car(c)==othw_ || car(c)==othw_append_ || car(c)==othw_replace_)

cell advance_pattern(cell c)

  {  /* c is a list pattern */
  if(car(c)==any_keyword_ || car(c)==anything_ || 
     car(c)==any_symbol_ || car(c)==any_binary_ ||
     car(c)==any_string_ || car(c)==any_list_ || car(c)==any_vector_ ||
     car(c)==any_float_ || is_list(car(c)) || is_vector(car(c))
     || car(c)==nothing_)
     {
     c = cdr(c);
     if(c!=NIL && cdr(c)!=NIL && is_otherwise(c))
       c = cddr(c);
     return c;
     }
  if(car(c)==the_symbol_ || car(c)==or_ || car(c)==and_ || car(c)==not_
     || car(c)==zero_more_ || car(c)==one_more_ || car(c)==set_of_ ||
     car(c)==sequence_ || car(c)==optional_)
     {
     c = cddr(c);
     if(c!=NIL && cdr(c)!=NIL && is_otherwise(c))
       c = cddr(c);
     return c;
     }
  if(is_cons(car(c)) || is_vector(car(c)))
    {
    c = cdr(c);
    if(c!=NIL && cdr(c)!=NIL && is_otherwise(c))
      c = cddr(c);
    return c;
    }
  if(car(c)==check_ || car(c)==rest_check_)
    {
    c = cdddr(c);
    if(c!=NIL && cdr(c)!=NIL && is_otherwise(c))
       c = cddr(c);
    return c;
    }
  slex("Internal il5check error: wrong ck clause \n    not detected by listcheck().");
  }

cell half_advance_pattern(cell c) /* just up to the possible otherwise_ */

  {  /* c is a list pattern */

  if(c==NIL || cdr(c)==NIL)
    return c;
  if(car(c)==any_keyword_ || car(c)==anything_ || 
     car(c)==any_symbol_ || car(c)==any_binary_ ||
     car(c)==any_string_ || car(c)==any_list_ || car(c)==any_vector_ ||
     car(c)==any_float_ || is_list(car(c)) || is_vector(car(c))
     || car(c)==nothing_)
     {
     c = cdr(c);
     return c;
     }
  if(is_cons(car(c)) || is_vector(car(c)))
    {
    c = cdr(c);
    return c;
    }
  if(car(c)==the_symbol_ || car(c)==or_ || car(c)==and_ || car(c)==not_
     || car(c)==zero_more_ || car(c)==one_more_ || car(c)==set_of_ ||
     car(c)==sequence_ || car(c)==optional_)
     {
     c = cddr(c);
     return c;
     }
  if(car(c)==check_ || car(c)==rest_check_)
    {
    c = cdddr(c);
    return c;
    }
  slex("Internal il5check error: wrong ck clause \n    not detected by listcheck().");
  }

cell combine_otws(cell pt, cell olrv)

  {
  if(pt==NIL || !is_otherwise(pt))
    return olrv;
  if(car(pt)==othw_)
    {
    if(!is_string(olrv))
      return cadr(pt);
    else
      return olrv;
    }
  if(car(pt)==othw_replace_)
    return cadr(pt);
  if(car(pt)==othw_append_ && is_string(olrv))
    {
    sprintf(il5cbuffer, "%s\n    %s",
      string(cadr(pt)), string(olrv));
    return strbuild(il5cbuffer);
    }
  if(car(pt)==othw_append_)
    return cadr(pt);
  return olrv;  /* in any other case */
  }


/* Data type used locally for checking sets */

struct setck {
  int flag;
  int must;
  cell item;
  };

/* Checking a list */

cell listcheck(cell what, cell pattern, int partial)

/* partial means that only the head of the list is checked */

  {
  cell rv;
  cell t;
  cell lwhat;
  int intr, i, p;
  int totl;

  struct setck *sc;

/* when occouring in a list calls are interpreted in a different way,
 * i.e. :or :and and :not can work with sequences and sequence
 * matchings are supported
 */

  rv = meta_syntax_error_;
  totl = 0;


  do {

   if(pattern==NIL && what!=NIL && partial)
      if(totl<10)
        return nr[totl];
      else
        return make(BINARY, (long)totl);

  	if(what==NIL && pattern!=NIL || what!=NIL && pattern==NIL)
      {
    	rv = syntax_error_;
      goto top_lev_err_;
      }

  if(car(pattern)==anything_ || car(pattern)==any_symbol_ ||
  car(pattern)==any_binary_ || car(pattern)==any_string_ ||
  car(pattern)==any_list_ || car(pattern)==any_vector_ ||
  car(pattern)==any_float_ || is_list(car(pattern)) || 
  car(pattern)==the_symbol_ ||
  is_vector(car(pattern)) || car(pattern)==check_ || car(pattern)==nothing_)
     	rv = check(car(what), pattern);
  	if(car(pattern)==not_)
    	{
    	rv = listcheck(what, cadr(pattern), 1);
    	if(!is_binary(rv)) 
      	rv = nr[0]; /* ! */
    	else
			rv = syntax_error_;
    	}
  	if(car(pattern)==and_)
     {
     intr = -1;
  	  for(t=cadr(pattern); t!=NIL; t = advance_pattern(t))
       {
       rv = listcheck(what, t, 1);
       if(is_binary(rv))
         {
         if(intr==-1 || binary(rv)<intr)
           intr = (int)binary(rv);
         }
       else
         {
         if((cddr(t)!=NIL) && caddr(t)==othw_ && !is_string(rv))
           return cadddr(t);   /* error is specified here */
         if((cddr(t)!=NIL) && caddr(t)==othw_replace_)
           return cadddr(t);   /* error is specified here */
         if((cddr(t)!=NIL) && caddr(t)==othw_append_)
           if(!is_string(rv))
             return cadddr(t);   /* error is specified here */
   		  else
             {
             sprintf(il5cbuffer, "%s\n    %s", string(cadddr(t)),
               string(rv));
             rv = strbuild(il5cbuffer);
             return rv;
             }
         goto top_lev_err_;
         }
       }
     if(intr<10)
		 rv = nr[intr];
     else
       rv = make(BINARY, (long)intr);
     }
   if(car(pattern)==or_)
     {
     intr = -1;
     for(t=cadr(pattern); t!=NIL; t = advance_pattern(t))
       {
       rv = listcheck(what, t, 1);
       if(is_binary(rv))
         break;  /* and rv remains the value */
       else
         {
         if((cddr(t)!=NIL) && caddr(t)==othw_ && !is_string(rv))
           return cadddr(t);   /* error is specified here */
         if((cddr(t)!=NIL) && caddr(t)==othw_replace_)
           return cadddr(t);   /* error is specified here */
         if((cddr(t)!=NIL) && caddr(t)==othw_append_)
           if(!is_string(rv))
             return cadddr(t);   /* error is specified here */
   		  else
             {
             sprintf(il5cbuffer, "%s\n    %s", string(cadddr(t)),
               string(rv));
             rv = strbuild(il5cbuffer);
             return rv;
             }
         goto top_lev_err_;
         }
       }
     if(!is_binary(rv))
       return rv; 
     }
   /* zero-or-more: the maximum nr. of items are checked */
   if(car(pattern)==zero_more_)
     {
     t = cadr(pattern);
     intr = 0;
     lwhat = what;
     do {
        rv = listcheck(lwhat, t, 1);
        if(!is_binary(rv))
          {
          if(intr>9)
            rv = make(BINARY, (long)intr);
          else
            rv = nr[intr];
          goto top_lev_err_;
          }
        intr += (int)binary(rv);
        for(i=0; i<binary(rv); i++)
          {
          if(lwhat==NIL)
            slex1i("Bad jump return (=%d) in argument to :zero-more.", 
                    (int)binary(rv));
          lwhat = cdr(lwhat);
          }
        }while(lwhat!=NIL);
     if(intr>9)
       rv = make(BINARY, (long)intr);
     else
       rv = nr[intr];
     }
   if(car(pattern)==optional_)
     {
     rv = listcheck(what, cadr(pattern), 1);
     if(!is_binary(rv))
        rv = nr[0];
     goto top_lev_err_;
     }
   if(car(pattern)==one_more_)
     {
     t = cadr(pattern);
     intr = 0;

     rv = listcheck(what, t, 1);
     if(!is_binary(rv))
       goto top_lev_err_;
     else
       intr = (int)binary(rv);
     
     lwhat = what;

     for(i=0; i<(int)binary(rv); i++)
       {
       if(lwhat==NIL)
         slex("Bad jump value returned by the first call for :one-more.");
       lwhat = cdr(lwhat);
       }

     do {
        rv = listcheck(lwhat, t, 1);
        if(!is_binary(rv))
          {
          if(intr>9)
            rv = make(BINARY, (long)intr);
          else
            rv = nr[intr];
          goto top_lev_err_;
          }
        intr += (int)binary(rv);
        for(i=0; i<binary(rv); i++)
          {
          if(lwhat==NIL)
            slex1i("Bad jump return (=%d) in argument to :one-more.", 
                    (int)binary(rv));
          lwhat = cdr(lwhat);
          }
        }while(lwhat!=NIL);

     if(intr>9)
       rv = make(BINARY, (long)intr);
     else
       rv = nr[intr];
     }
   if(car(pattern)==set_of_)
     {
     intr = 0;
     for(p = 0, t = cadr(pattern); t!=NIL; t = advance_pattern(t))
       p ++;
     sc = malloc(sizeof(struct setck)*p);
     if(sc==NULL)
       slex("No more memory during listcheck().");
     for(i=0, t = cadr(pattern); t!=NIL; t = advance_pattern(t), i++)
       {
       if(car(t)==required_)
         {
         sc[i].must = 1;
			t = cdr(t);
         }
       else
         sc[i].must = 0;
       sc[i].flag = 0;  /* not yet used */
       sc[i].item = t;
       }
     lwhat = what;
     while(lwhat!=NIL)
       {
       rv = NIL;
       for(i=0; i<p; i++)
         {
         if(sc[i].flag)
           continue;
         rv = listcheck(lwhat, sc[i].item, 2); /* only check the first item */
         if(!is_binary(rv))
           continue;
         intr += (int)binary(rv);
         sc[i].flag = 1;
         break;
         }
       if(!is_binary(rv))  /* none more was found */
         {
         il5cbuffer[0] = 0;
         rv = NIL;
         for(i=0; i<p; i++)
           {
           if(!sc[i].flag && sc[i].must)
             {
             /* It is some error */
             rv = syntax_error_;
             t = half_advance_pattern(sc[i].item);
             if(is_otherwise(t))
               sprintf(il5cbuffer + strlen(il5cbuffer),
                 "%s", string(cadr(t)));
             }
           }
         if(rv==NIL)  /* no error detected */
            {
/*            printf("set-of succeeded, intr=%d\n", intr); */
         	if(intr<10)
           	  rv = nr[intr];
         	else
           	  rv = make(BINARY, (long)intr);
            break;  /* search finished with success */
            }
			if(rv==syntax_error_)
           {
           if(il5cbuffer[0])
             rv = strbuild(il5cbuffer);
           break;
           }
         }
       for(i=0; i<binary(rv); i++)
         lwhat = cdr(lwhat);
       }
     free(sc);
     }
   if(car(pattern)==sequence_)
     {
     rv = listcheck(what, cadr(pattern), 1);
     }
   if(car(pattern)==rest_check_)
     {
     /* The function itself OR it's name is expected there */
     rv = cadr(pattern);
     if(is_symbol(rv))
       rv = symval(rv);
     push(what);
     push(caddr(pattern));
     rv = appl(rv, 2);   
     }

top_lev_err_:
  	
   if(rv==meta_syntax_error_)
    	return rv;

  	if(!is_binary(rv))
    	{ 
      rv = combine_otws(half_advance_pattern(pattern), rv);
      return rv;
    	}

   totl += (int)binary(rv);

   if(partial==2)  /* just for set_of_ */
     {
     if(totl<10)
       return nr[totl];
     else
       return make(BINARY, (long)totl);
     }

/* Advancing pattern and what */

   for(i=0; i<(int)binary(rv); i++)
      {
      if(what==NIL)
         slex1i("Bad jump return (=%d) in check clause.", (int)binary(rv));
      else
         what = cdr(what);
      }

   pattern = advance_pattern(pattern);

/* Terminating a correct partial construction */

   if(partial && pattern==NIL)
     {
     if(totl<10)
       return nr[totl];
     else
       return make(BINARY, (long)totl);
     }

   if(what==NIL && pattern==NIL)
      {
      return nr[1];
      }

  	 } while(1);
  }

cell vectorcheck(cell what, cell pattern)

  {
  slex("vector check not implemented in this beta version.");
  }




/* Checking some scalar construction */

cell check(cell what, cell pattern)

  {
  cell rv;
  cell t;

  rv = meta_syntax_error_;

/* The names of metapat constructions cannot be keywords! */

  while(is_symbol(car(pattern)) && !is_keyword(car(pattern)))
    pattern = lappend(t = symval(car(pattern)), cdr(pattern));

  if(t==no_sym_val_)
    slex1s("No symbol named %s containing a pattern.\n");

/* TAKE CARE: this can lead to unreadable syntax macros! */

  if(is_cons(car(pattern)))
    {
    if(is_cons(what))
      rv = listcheck(what, car(pattern), 0);
    else
      rv = syntax_error_;
    }
  if(is_vector(car(pattern)))
    {
    if(is_vector(what))
      rv = vectorcheck(what, car(pattern));
    else
      rv = syntax_error_;
    }

  if(car(pattern)==zero_more_)
    slex(":zero-or-more can only appear in a list or vector.");
  if(car(pattern)==one_more_)
    slex(":one-or-more can only appear in a list or vector.");
  if(car(pattern)==set_of_)
    slex(":set-of can only appear in a list or vector.");
  if(car(pattern)==sequence_)
    slex(":sequence can only appear in a list or vector.");
  if(car(pattern)==rest_check_)
    slex(":rest_check_ can only appear in a list.");  
      /* NOT IN A VECTOR */

  if(car(pattern)==nothing_) 
    return nr[0];
  if(car(pattern)==anything_) 
    return nr[1];
  if(car(pattern)==any_symbol_)
    if(is_symbol(what))
      return nr[1];
    else 
      rv = syntax_error_;
  if(car(pattern)==any_keyword_)
    if(is_keyword(what))
      return nr[1];
    else 
      rv = syntax_error_;
  if(car(pattern)==any_binary_)
    if(is_binary(what))
      return nr[1];
    else 
      rv = syntax_error_;
  if(car(pattern)==any_string_)
    if(is_string(what))
      return nr[1];
    else 
      rv = syntax_error_;
  if(car(pattern)==any_list_)
    if(is_list(what))
      return nr[1];
    else 
      rv = syntax_error_;
  if(car(pattern)==any_vector_)
    if(is_vector(what))
      return nr[1];
    else 
      rv = syntax_error_;
  if(car(pattern)==any_float_)
    if(is_float(what))
      return nr[1];
    else 
      rv = syntax_error_;
  if(car(pattern)==the_symbol_)
    if(is_symbol(what) && what==cadr(pattern))
      return nr[1];
    else 
      rv = syntax_error_;
  if(car(pattern)==or_)
    {
    for(t=cadr(pattern); is_cons(t); t = advance_pattern(t))
      {
      rv = check(what, t);
      if(is_binary(rv))
        return nr[1];  
      }
         /* as only single item can be checked for here,
          * sequences not, under any circumstances
          */
    rv = syntax_error_;
    }
  if(car(pattern)==and_)
    {
    for(t=cadr(pattern); is_cons(t); t = advance_pattern(t))
      {
      rv = check(what, t);
      if(!is_binary(rv))  /* i.e. this is an error */
        {
        t = half_advance_pattern(t);
        rv = combine_otws(t, rv);
        break;
        }
      }
    if(is_binary(rv))
      return nr[1];  /* one item checked */
         /* as only single item can be checked for here,
          * sequences not, under any circumstances
          */
    }
  if(car(pattern)==not_)
    if((rv = check(what, cadr(pattern)))==NIL)
      rv = syntax_error_;
    else
      return nr[1];
  if(car(pattern)==check_)
     {
     rv = cadr(pattern);
     if(is_symbol(rv))
       rv = symval(rv);
     push(what);
     push(caddr(pattern));
     rv = appl(rv, 2);   
     if (is_binary(rv)) 
       return rv;
     }
  if(rv==meta_syntax_error_)
    return rv;
  if(!is_binary(rv))
    {
    pattern = half_advance_pattern(pattern);
    rv = combine_otws(pattern, rv);
    return rv;
    }
  return rv;		
  }

#ifdef ANIMATED

int thecheck(int n)

  {
  cell o;
  cell p;

  if(n==-1)
    n = 2;
  if(n!=2)
    slex("check syntax: (ckeck object pattern).");
  p = pop();
  o = pop();
  push(check(o,p));
  return 1;
  }
#endif

il5check()

  {
  int i;

  anything_ = esym(":anything");
  nothing_ = esym(":nothing");
  any_symbol_ = esym(":any-symbol");
  any_keyword_ = esym(":any-keyword");
  any_binary_ = esym(":any-binary");
  any_string_ = esym(":any-string");
  any_list_ = esym(":any-list");
  any_vector_ = esym(":any-vector");
  any_float_ = esym(":any-float");
  the_symbol_ = esym(":the-symbol");
  or_ =       esym(":or");
  and_ =      esym(":and");
  zero_more_ = esym(":zero-more");
  one_more_ = esym(":one-more");
  not_ =     esym(":not");
  set_of_ = esym(":set-of");
  sequence_ = esym(":sequence");
  check_ = esym(":check");
  rest_check_ = esym(":rest-check");
  optional_ = esym(":optional");
  required_ = esym(":required");
  othw_ = esym(":otherwise");
  othw_replace_ = esym(":otherwise-replace");
  othw_append_ = esym(":otherwise-append");

  syntax_error_ = esym(":syntax-error");
  meta_syntax_error_ = esym(":meta-syntax-error");

  for(i=0; i<10; i++)
    nr[i] = make(BINARY, (long)i);

#ifdef ANIMATED
  bind(esym("check"), makf(thecheck));
#endif

  }

/* END OF THE IL4CHECK MODULE */


