/*    ILFIVE lisp objects library -- packages
 *
 *    Copyright (c) 1991--1998 Alexandru Dan Corlan M.D.
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    In short, you can use and modify this program. You can
 *    distribute the program with or without your modifications if you want,
 *    but you must not require licence fees and must provide the
 *    full source code with the distribution. If you make an application program
 *    which just calls functions from this library you may distribute that
 *    application program without source code, but also as an linkable object file, 
 *    and must distribute also this library with source code.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <malloc.h>
#include <stdio.h>

#include <shpell5.h>

/* symbols that define the way the hash tables work
 * in this subsystem
 */

#define START_SIZE 10101
#define START_INCR 1131
#define START_IINCR 111
#define THRESHOLD 127

cell AllPacks;
cell crtpack;

/* The standard packages */

cell _Keyword;
cell _User;
cell _Lisp;
cell _System;

/*    THE C-only usable section   */

cell _make_pack(char *pn)

  {
  cell rv;
  struct paks *pks;
  struct hent *htable;
  register int i;

  /* The list of all packages is not checked */

  pks = (struct paks *)alloc(sizeof(struct paks));
  htable = (struct hent *)alloc(sizeof(struct hent) * START_SIZE);
  if (pks==NULL || htable==NULL)
    {
    slex("There isn't enough memory to build package.");
    }
  rv = lmake(PACKAGE);
  packname(rv) = cons(strbuild(pn), NIL);
  packstrc(rv) = pks;
  pks->usel = NIL;
  pks->incr = START_INCR;
  pks->iincr = START_IINCR;
  pks->htsz = START_SIZE;
  pks->thres = THRESHOLD;
  pks->htab = htable;
  for (i=0; i<START_SIZE; i++)
    {
    htable[i].heflags = 0;
    htable[i].hecont = NO_CELL;  /* no symbol in fact */
    }

  /* The package is now inserted in the package list */

  AllPacks = cons(rv, AllPacks);
  return rv;
  }

/* The following function takes a string
 * and returns the associated package
 */

static int _is_the_pack(cell pack, char *pn)

  {
  register cell cl;

  cl = packname(pack);
  while(cl!=NIL)
    {
    if (!strcmp(pn, string(car(cl))))  /* case is observed */
      return 1;  /*, yes, it is the name or nickname of pack */
    cl = cdr(cl);
    }
  return 0;  /* no */
  }

cell _find_pack(char *pn)  /* be it a name or a nickname */

  {
  register cell pk;

  pk = AllPacks;
  while(pk!=NIL)
    {
    if(_is_the_pack(car(pk), pn) != 0)
      {
      return car(pk);
      }
    pk = cdr(pk);
    }
  return NIL;  /* couldn't be found */
  }

cell _in_pack(char *pn)

  {
  cell p;

  p = _find_pack(pn);
  if (p!=NIL)
    {
    crtpack = p;
    return p;
    }
  crtpack = p = _make_pack(pn);
  return p;
  }


char *_pack_name(cell pk)

  {
  return string(car(packname(pk)));
  }

cell pack_nn(cell pk)

  {
  return cdr(packname(pk));
  }

/* Rename package is not yet implemented */

cell _pk_use_l(cell pk)

  {
  return packstrc(pk)->usel;
  }

cell _pk_use_bl(cell pk)

  {
  register cell pl;
  register cell rv;
  register cell upl;

  rv = NIL;
  pl = AllPacks;
  while(pl!=NIL)
    {
    upl = packstrc(car(pl))->usel;
    while(upl!=NIL)
      {
      if (car(upl)==pk) rv = cons(car(pl), rv);
      upl = cdr(upl);
      }
    pl = cdr(pl);
    }
  return rv;  /* naturally it may be NIL */
  }

cell _p_shing_sy(cell pk)

  {
  register cell rv;
  register i;

  rv = NIL;
  for(i=0; i<(packstrc(pk)->htsz); i++)
    {
    if(is_shdowing(packstrc(pk)->htab[i]))
      rv = cons(car(packstrc(pk)->htab[i].hecont), rv);
    }
  return rv;
  }

cell list_all_p(void)

  {
  return AllPacks;
  }

cell j_makesymbol(char *sn)

  {
  register cell rv;

  sn = name(sn);
  rv = givecell();
  id(rv) = SYMBOL;
  symname(rv) = sn;
  sympack(rv) = NIL;
/*  printf("<<%s>> ", sn); */

  return rv;
  }

int packhash(char *str, int n)

  {
  int hv;
  int i;

  hv = 0;
  while(*str)
    {
    for (i=0; i<6; i++)
      hv = (hv * 2 + *str) & 0x7fff;
    str++;
    }
  return hv % n;
  }

static void htenlarge(cell pk)

  {
  struct hent *old, *new;
  int oldsz, newsz;
  int i;
  int oldi, newi;

  newsz = (oldsz = packstrc(pk)->htsz) + packstrc(pk)->incr;
  packstrc(pk)->incr += packstrc(pk)->iincr;
  printf("Enlarging %s from %d to %d entries.\n",
	 string(car(packname(pk))), oldsz, newsz); 
  fflush(stdout); 
  new = (struct hent *)alloc((sizeof(struct hent))*newsz);
  old = packstrc(pk)->htab;

  /* clear all the new entries */
  for(newi = 0; newi < newsz; newi++)
    {
    new[newi].heflags = 0;  /* free */
    new[newi].hecont = NIL;
    }
  /* rehash all the old entries in new places;
   * NOTICE: as incr was incremented as a side-effect until
   * now, it is possible to recall this same function
   * recursivelly if the rehash still fails, and in this case
   * the atempt is done with a larger increment (a real chance)
   * and so on.
   */
  for(oldi = 0; oldi < oldsz; oldi++)
    {
    if(is_free(old[oldi])) continue;
    if(is_symbol(old[oldi].hecont))
      newi = packhash(symname(old[oldi].hecont), newsz);
    else
      newi = packhash(symname(car(old[oldi].hecont)), newsz);
    for(i=0; i<packstrc(pk)->thres; i++)
      if (is_free(new[(newi+i)%newsz]))
        {
        new[(newi+i)%newsz].heflags = old[oldi].heflags;
        new[(newi+i)%newsz].hecont = old[oldi].hecont;
        break;
        }
    if (i>=packstrc(pk)->thres)
      {
      mfree(new);  /* not needed any more */
      printf("RENLARGE: ");
      fflush(stdout);
      htenlarge(pk);
      return ;
      }
    }  /* very good, rehashing was finished here */
  packstrc(pk)->htsz = newsz;
  packstrc(pk)->htab = new;
  mfree(old);
  return ; /* graciously */
  }


int _2nd_findsym;
int _2nd_intern;

static cell __finds(char *str, cell pk)

  {
  int i, hsz, t;
  register cell vl;

  _2nd_findsym = _NONE;
  t = packhash(str, hsz = packstrc(pk)->htsz);
  for(i=0; i<(packstrc(pk))->thres; i++)
    {
    vl = packstrc(pk)->htab[(t+i)%hsz].hecont;
    if(vl == NIL) break;  /* not found here */
    if(is_symbol(vl) && !strcmp(symname(vl),str))
      {
      _2nd_findsym =
        is_external(packstrc(pk)->htab[(t+i)%hsz]) ? _EXTERNAL : _INTERNAL;
      return vl;
      }
    if(is_list(vl) && !strcmp(symname(car(vl)), str))
      {
      _2nd_findsym =
        is_external(packstrc(pk)->htab[(t+i)%hsz]) ? _EXTERNAL : _INTERNAL;
      return car(vl);
      }
    }
  return NIL;
  }


cell _findsym(char *str, cell pk)

  {
  cell rv;
  cell pl;

  _2nd_findsym = _NONE;
  if ((rv = __finds(str,pk)) != NIL)
    return rv;  /* and _2nd_find is set O.K. */
  /* in this case try each of the used packages */
  pl = packstrc(pk)->usel;
  while(pl!=NIL)
    {
    if((rv = __finds(str,car(pl)))!=NIL)
      {
      if(_2nd_findsym == _EXTERNAL)
        {
        _2nd_findsym = _INHERITED;
        return rv;
        }
      }
    pl = cdr(pl);
    }
  _2nd_findsym = _NONE;
  return NIL;  /* none could be found */
  }

cell _intern(char *str, cell pk)

  {
  register cell rv;
  register i, t, hsz;

  if ((rv=_findsym(str,pk))!=NIL)
    {
    _2nd_intern = _2nd_findsym;
    return rv;
    }
  rv = j_makesymbol(str);
  sympack(rv) = pk;  /* this becomes the home package of the symbol */
  do {
     t = packhash(str, hsz = packstrc(pk)->htsz);
     for (i=0; i<packstrc(pk)->thres; i++)
       {
       if (is_free(packstrc(pk)->htab[(i+t)%hsz]))
         {
         if (pk == _Keyword)
           /* i.e. it must be made external */
           {
           _2nd_intern = _EXTERNAL;
           packstrc(pk)->htab[(i+t)%hsz].heflags = EXTERN_HMASK | NOT_FREE_HMASK;
           }
         else
           {
           packstrc(pk)->htab[(i+t)%hsz].heflags = NOT_FREE_HMASK;
           _2nd_intern = _INTERNAL;
           }
         packstrc(pk)->htab[(i+t)%hsz].hecont = rv;
         return rv;
         }
       }
     htenlarge(pk);
     } while(1);  /* i.e. until enough space is found */
   /* this place is never reached */
   }

int _unintern(cell sy, cell pk)

  {
  struct hent *ht;
  register int i;
  cell prev;

  ht = packstrc(pk)->htab;
  for (i=0; i<packstrc(pk)->htsz; i++)
    {
    if (is_free(ht[i])) continue;
    if (is_symbol(ht[i].hecont))
      {
      /* removes a non-shadowing symbol */
      if (ht[i].hecont!=sy) continue;
      ht[i].hecont = NIL;
      ht[i].heflags = 0;
      if (sympack(sy)==pk)
        sympack(sy) = NIL;
      return 1;   /* a symbol WAS actually removed */
      }
    if (is_list(ht[i].hecont))
      {
      /* removes a shadowing symbol */
      if (car(ht[i].hecont)!=sy) continue;
      prev = ht[i].hecont;
      ht[i].hecont = cdr(prev);
      ungcell(prev);
        /* never accessed from somewhere else
         * as this way of implementing things is hidden
         */
      if(ht[i].hecont==NIL)
        ht[i].heflags = 0;
      else
        if(cdr(ht[i].hecont)==NIL) /* it doesn't shadow any more */
          {
          prev = ht[i].hecont;
          ht[i].hecont = car(prev);
          ungcell(prev);
            /* never accessed from somewhere else
             * as this way of implementing things is hidden
             */
           ht[i].heflags &= !SHADOW_HMASK;
           }
       /* symbol removed */
       if (sympack(sy)==pk)
         sympack(sy) = NIL;
       return 1;   /* a symbol WAS actually removed */
       }
     }
   return 0;  /* none could be removed */
   }

void _exort(cell slors, cell pk)

  {
  register i, hsz, htv;
  struct hent *hen;
  cell vl;

  if (is_list(slors))
    {
    while(slors!=NIL)
      {
      _exort(car(slors), pk);
      slors = cdr(slors);
      }
    return ;
    }
  /* slors is now expected to be a symbol */
  htv = packhash(symname(slors), hsz = packstrc(pk)->htsz);
  hen = packstrc(pk)->htab;
  for(i=0; i<packstrc(pk)->thres; i++)
    {
    vl = hen[(htv+i)%hsz].hecont;
    if (is_symbol(vl) && vl == slors ||
        is_list(vl) && car(vl) == slors)
        {
        hen[(htv+i)%hsz].heflags = hen[(htv+i)%hsz].heflags | EXTERN_HMASK;
        return ;
        }
    }
  return ;  /* not found, thus not exported */
  }


void _unexport(cell slors, cell pk)

  {
  register i, hsz, htv;
  struct hent *hen;
  cell vl;

/* An error should be issued here if it is the keyword:
 * package
 */

  if (is_list(slors))
    {
    while(slors!=NIL)
      {
      _unexport(car(slors), pk);
      slors = cdr(slors);
      }
    return ;
    }
  /* slors is now expected to be a symbol */
  htv = packhash(symname(slors), hsz = packstrc(pk)->htsz);
  hen = packstrc(pk)->htab;
  for(i=0; i<packstrc(pk)->thres; i++)
    {
    vl = hen[(htv+i)%hsz].hecont;
    if (is_symbol(vl) && vl == slors ||
        is_list(vl) && car(vl) == slors)
        {
        hen[(htv+i)%hsz].heflags = hen[(htv+i)%hsz].heflags & ~EXTERN_HMASK;
        return ;
        }
    }
  return ;  /* not found, thus not exported */
  }

void _use(cell pktouse, cell pk)

  {
  if (!memq(pktouse, packstrc(pk)->usel))
    packstrc(pk)->usel = cons(pktouse, packstrc(pk)->usel);
  return ;
  }

/* The following functions are traditionally used
 * by Il packages, so it seems irrational not to
 * include them here
 */

cell esym(char *snm)   /* adds a symbol to the current package */

  {
  cell rv;

  if (*snm!=':')
    {
    rv = _intern(snm, crtpack);
    _exort(rv, crtpack);
    }
  else
    {
    rv = _intern(snm+1, _Keyword);
    _exort(rv, _Keyword);
    }
  return rv;
  }

cell emakesym(char *snm, cell pack)

  {
  cell rv;

  rv = _intern(snm, pack);
  _exort(rv, pack);
  return rv;
  }


cell pmakesym(char *snm, char *pn)

  {
  cell rv;
  cell pack;

  if ((pack =_find_pack(pn)) == NIL)
    slex1s("Package %s is specified but wasn't ever built.", pn);
  rv = _intern(snm, pack);
  _exort(rv, pack);
  return rv;
  }

void il5packs(void)   /* initialises a bunch of standard packages */

  {
  _Keyword = _in_pack("KEYWORD");
  _Lisp = _in_pack("LISP");
  _System = _in_pack("SYSTEM");
  _User = _in_pack("USER");
  _in_pack("USER");
  _use(_Lisp,_User);
  _use(_Lisp,_System);
  _use(_System,_User);
  Nil = emakesym("nil", _Lisp);
  t_ = emakesym("t", _Lisp);
  }


void il5god(void)

  {
  AllPacks = crtpack = NIL;
  }


 
