/*    ILFIVE lisp objects library -- hash tables
 *
 *    Copyright (c) 1991--1998 Alexandru Dan Corlan M.D.
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    In short, you can use and modify this program. You can
 *    distribute the program with or without your modifications if you want,
 *    but you must not require licence fees and must provide the
 *    full source code with the distribution. If you make an application program
 *    which just calls functions from this library you may distribute that
 *    application program without source code, but also as an linkable object file, 
 *    and must distribute also this library with source code.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include <shpell5.h>

long lochash(cell obj)

  {
  long rv;
  char *chp;
  int i;

  if (obj==NIL) return 1;
  switch(id(obj))
    {
    case PACKAGE:
    case HASHTABLE:
    case SYMBOL:
      return ((long)obj & 0x7fffffffL);
    case BINARY:
      return (binary(obj) & 0x7fffffffL);
    case STRING:
    case PURESTRING:
      for(i=0, rv = 0L, chp = string(obj); *chp; chp++)
        rv += *chp * (i++);
      rv = rv & 0x7fffffffL;
      return rv;
    case FLOAT:
      rv = ((long)(car(obj)) + (long)(cdr(obj)));
      rv = rv & 0x7fffffffL;
      return rv;
    case COMPLEX:
      return lochash(real(obj))/2 + lochash(imag(obj))/2;
    default:
      return 1;
    }
  return 1;
  }

long locshash(cell obj)

  {
  if (is_cons(obj))
    return lochash(car(obj)) + locshash(cdr(obj));
  return lochash(obj);
  }

int cellhash(cell obj, int size)

  {
  return ((locshash(obj) & 0x7fffL) % size);
  }

cell makeht(cell test, int sz, int incr, int thr)

  {
  struct hashtab *thet;
  cell rv;
  register i;

  /* check here if test is eq, eql or equal */

  thet = (struct hashtab *)alloc((long)sizeof(struct hashtab));
  if (thet==NULL)
    slex("Not enough memory for hash table.");
  thet->size = sz;
  thet->increment = incr;
  thet->threshold = thr;
  thet->tab = (struct hashent *)alloc((long)sizeof(struct hashent) * sz);
  for(i=0; i<sz; i++)
    thet->tab[i].key = thet->tab[i].value = NIL;  /* this might raise some problems */
  rv = lmake(HASHTABLE);
  matchtype(rv) = test;  /* provided test is eq, eql or equal */
  htobj(rv) = thet;
  return rv;
  }

int jgethash(cell k, cell ht)

  {
  struct hashent *he;
  int j, i;
  int rv;
  int sz;

  rv = -1;

  he = httab(ht);
  sz = htsize(ht);
  for (i = cellhash(k, htsize(ht)), j = 0; j < htthre(ht); j++)
    {
    if (is_eq(matchtype(ht), k, he[(i+j) % sz].key))
/*    if(k==he[(i+j) % sz].key) */
      {
/*      print(he[(i+j) % sz].key, stdout);*/
/*      print(k, stdout);*/
/*      printf("RETFOUND: %d %d ", i, (i+j) % sz);*/
      return (i+j) % sz;
      }
    if (he[(i+j) % sz].key == NIL && he[(i+j) % sz].value==NIL)
      if (rv==-1) rv = (i+j) % sz;  /* search continues as this may be a remhashed entry */
    }
/*  printf("RET: %d %d ", i, rv);*/
  return rv;
  /* -1 if no place for the value exists in the table and
   * the entry where place was found if that was present
   */
  }

cell hashretrieve(cell ht, cell k, cell def)

  {
  int p;

  if((p = jgethash(k, ht))==-1 ||
     (htval(ht,p)==NIL) && (htkey(ht,p)==NIL))
    return def;  /* return the default value */
  return htval(ht,p);
  }

cell rehash(cell ht)

  {
  int p;
  static nht;
  struct hashent *newhe;
  int newsize;
  int i, j, newincr;
  int impossiblef;  /* it is impossible to rehash it, try again */

  /* the table is rehashed */
  /* 1) allocate the new table */
  printf("Rehashing some table, [%d %d %d].\n",
    htsize(ht), htincr(ht), htthre(ht));
  newsize = htsize(ht) + htincr(ht);
  if(newsize>4095)
    slex("Implementation cannot manage so large a hash table.");
  newincr = (int) (((long)htincr(ht) * INCRINCR)/100 + 1)
            + htincr(ht);
  newhe = (struct hashent *)alloc((long)sizeof(struct hashent) * newsize);
  /* clear it */
  for(i=0; i<newsize; i++)
    newhe[i].key = newhe[i].value = NIL;
  /* rehash all the entries */
  for(i=0; i<htsize(ht); i++)
    if(htkey(ht,i)!=NIL && htval(ht,i)!=NIL)
      {
      p = cellhash(htval(ht,i), newsize);
      impossiblef = 1;
      for(j=0; j<htthre(ht); j++)
        if(newhe[(p+j) % newsize].key==NIL && newhe[(p+j) % newsize].value==NIL)
          {
          newhe[(p+j) % newsize].key = htkey(ht,i);
          newhe[(p+j) % newsize].value = htval(ht,i);
          impossiblef = 0;
          break;
          }
      if(impossiblef)
        {
        mfree(newhe);
        htincr(ht) = newincr;
        return rehash(ht);
        }
      }
  mfree(httab(ht));
  htincr(ht) = newincr;
  htsize(ht) = newsize;
  httab(ht) = newhe;
  return ht;
  }


cell hashstore(cell ht, cell k, cell v)

  {
  int p;

  if((p = jgethash(k, ht))==-1)
    {
    rehash(ht);
    hashstore(ht, k, v);
    return v;
    }
  if(htkey(ht,p)==NIL)
    htkey(ht,p) = k;
  htval(ht,p) = v;  /* the old value not affected by this rplaca */
  return v;
  }

cell dhashstore(cell ht, cell k, cell v)

  {
  int p;

  if((p = jgethash(k, ht))==-1)
    {
    rehash(ht);
    hashstore(ht, k, v);
    return v;
    }
  if(htkey(ht,p)==NIL)
    htkey(ht,p) = k;
  if(htval(ht,p)!=NIL)
    ungstruct(htval(ht,p));
  htval(ht,p) = v;  /* the old value not affected by this rplaca */
  return v;
  }


place gethash(cell ht, cell k)

  {
  int p;

  /* a place for k must always exist */
  if((p = jgethash(k, ht))==-1)
    {
    rehash(ht);
    hashstore(ht, k, NIL);
    p = jgethash(k, ht);  /* if p==-1 it is "impossible value" */
    if(p==-1)
      slex("Impossible value returned by internal call to jgethash.");
    }
  if(htkey(ht,p)==NIL)
    htkey(ht,p) = k;
  return &htval(ht,p);
  }

int remhash(cell k, cell ht)

  {
  int p;

  p = jgethash(k, ht);
  if(p==-1) return 0;  /* no entry */
  htkey(ht,p) = htval(ht,p) = NIL;
  return 1;  /* entry deleted */
  }


/* End of the hashtable data type */
 
