/*    ILFIVE lisp objects library -- streams
 *
 *    Copyright (c) 1991--1998 Alexandru Dan Corlan M.D.
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    In short, you can use and modify this program. You can
 *    distribute the program with or without your modifications if you want,
 *    but you must not require licence fees and must provide the
 *    full source code with the distribution. If you make an application program
 *    which just calls functions from this library you may distribute that
 *    application program without source code, but also as an linkable object file, 
 *    and must distribute also this library with source code.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
\title{ILFOUR --- streams in/out package (il5io)}
\author{Alexandru Dan CORLAN}
\date{6th of July 1991}

\begin{abstract}
This file implements streams. A  \hlink{stream} is a 
\hlink{cell} of \hlink{id} \hlink{STREAM} i.e. on which
stream primitives and macros apply. Stream can be either
for input, either for input/output or both, and they can
write in files, windows, strings and in other streams,
and read from similar devices. There are two kinds of
functions that apply to them: control functions and i/o
functions; the i/o functions are character i/o (putc, getc, ungetc
but eventually called from print, read and format) and
virtual i/o (as are for example graphic primitives).

A supplementary attribute is associated to streams for the
purpose of graphic i/o, i.e. the way in which graphic
primitives are translated to output. There are in this
respect graphic windows, \hlink{PostScript} files and \hlink{TeX} files
(in which graphics are written as \hlink{ANTeX} descriptions).
The same graphic description as described by graphic primitives
is output in similar ways in a graphic window, a PostScript
or an ANTeX file.

A supplementary attribute is associated to input window stream
for the purpose of implementing edited input. There are two
levels of edited input: line-edit and history-edit.
\end{abstract}


 */

#include <malloc.h>
#include <stdio.h>

#include <shpell5.h>
#include <il5vars.h>

#undef USERBUFSIZE
#define USERBUFSIZE 256


/* \section{stream connection and disconnection} */

/* \notion{strconnect} */

cell strconnect(int obj_type, void *obj_description, int open_mode)

  {
  cell rv;

  rv = lmake(STREAM);
  streamid(rv) = obj_type;
  thestream(rv) = malloc(sizeof(CSTREAM));
  stream(rv)->styp = obj_type;
  stream(rv)->iotyp = open_mode;
  stream(rv)->exte = obj_description;
  stream(rv)->crmode = 1;  /* connected, not opened */
  stream(rv)->userinbuf = NULL;
  stream(rv)->userincrs = 0;
  switch(obj_type)
    {
    case MEMORY: 
      stream(rv)->size = strlen(obj_description);
      stream(rv)->fill = 0;
      break;
    case CHAR_FILE:
    case BINARY_FILE:
      stream(rv)->size = 0;  /* not known */
      stream(rv)->fill = ftell(obj_description);  /* expected to be a FILE * */
      break;
    case KEYBOARD:
    case SINK:
      stream(rv)->size = stream(rv)->fill = 0;
      break;
    default:
      slex1i("Invalid file type, %d, given argument to strconnect.", obj_type);
    }
  stream(rv)->hist = NIL;
  stream(rv)->ugstop = 0;
  stream(rv)->eoflag = 0;
  stream(rv)->line_nb = 0;
  stream(rv)->unclipst = NIL;
  return rv;
  }

/* \notion{disconnect} */

void *disconnect(cell sr)

  {
  void *rv;
  int i;

  switch(streamid(sr))
    {
    case KEYBOARD:
    case SINK:
      rv = NULL;
      break;
    case CHAR_FILE:
    case BINARY_FILE:
      for(i=stream(sr)->ugstop-1; i>=0; i--)
        ungetc(stream(sr)->ugstack[i], file_of(sr));
      /* NO break */
    case MEMORY:
      rv = stream(sr)->exte;
      break;
    default:
      slex1i("Cannot disconnect a stream of type %d.", (int)streamid(sr));
    }
  free(stream(sr));
  return rv;
  }

/* \section{describing streams} */

/* These functions specify more features of streams
   then what was said at connection
 */

/* \notion{dcr_mem_size}

This function, given a \hlink{string stream} specifies the
size of the associated piece of memory in bytes.
*/

void dcr_mem_size(cell sr, long sz)

  {
  if(streamid(sr)!=MEMORY)
    slex("dcr_mem_size() was called for a non-memory stream.");
  stream(sr)->size = sz;
  }

/* \section{Opening streams} 
   This section contains functions that open external i/o
   environments and than link streams to them. 
   The opened file is returned.
*/

cell fileopen(char *filename, int mode, int form)

  {
  FILE *f;
  char *fm;
  cell rv;

/*  fprintf(stderr, "Opening |%s|\n", filename);*/
  if(form!=BINARY_FILE)
    {
  	 if(mode==INPUT)
    	fm = "r";
  	 if(mode==OUTPUT)
    	fm = "w";
  	 if(mode==INPUT_OUTPUT)
    	fm = "r+";
    }
  else
    {
  	 if(mode==INPUT)
    	fm = "rb";
  	 if(mode==OUTPUT)
    	fm = "wb";
  	 if(mode==INPUT_OUTPUT)
    	fm = "r+b";
    }
  f = fopen(filename, fm);
  if(f==NULL)
    slex1s("Cannot open file %s in fileopen().\n", filename);
  rv = strconnect(form, f, mode);
  stream(rv)->crmode = 0;
  return rv;
  }

/* \subsection{the main stream getchar function} */

int srgetch(cell sr)

  {
  int r;

  if(stream(sr)->iotyp==OUTPUT)
    slex("Cannot read from an output stream.");

  if(stream(sr)->ugstop)
    {
    stream(sr)->ugstop --;
    stream(sr)->fill ++;
    return stream(sr)->ugstack[stream(sr)->ugstop];
    }
  switch(streamid(sr))
    {
    /* link streams not yet implemented */
    case KEYBOARD:
      stream(sr)->fill ++;
      return getc(stdin);
/* return extgetch was here before */
    case MEMORY:
      if((r = string_of(sr)[stream(sr)->fill]))
        {
        stream(sr)->fill ++;
        if(r=='\n')
          stream(sr)->line_nb ++;
        return r;
        }
      else
        return EOF;
    case CHAR_FILE:
    case BINARY_FILE:
      r = getc(file_of(sr));
      if(r=='\n')
        stream(sr)->line_nb ++;
      stream(sr)->fill ++;
      return r;
    default:
      slex("Only direct i/o slots can be read for the moment.");
    }
  }


void srputc(int c, cell sr)

  {
  if(stream(sr)->iotyp==INPUT)
    slex("Cannot print to an input stream.");
  switch(streamid(sr))
    {
    case CHAR_FILE:
    case BINARY_FILE:
      putc(c, file_of(sr));
      stream(sr)->fill ++;
      return;
    case MEMORY:
      if(stream(sr)->fill >= stream(sr)->size-1)
        slex("String filled up in output function.");
      string_of(sr)[stream(sr)->fill] = c;
      stream(sr)->fill ++;
      string_of(sr)[stream(sr)->fill] = 0; 
      return;
    case TeX_FILE:
    case PS_FILE:
      slex("TeX and PostScript not yet implemented.");
    default:
      slex1i("Output to streams of type %d not implemented yet.",
	     (int)streamid(sr));
    }
  }

int srungetch(int c, cell sr)

  {
  if(stream(sr)->iotyp==OUTPUT)
    slex("Cannot unget characters in output streams.");
  if(stream(sr)->ugstop >= LARGEST_OP)
    slex("Stream has the ungetc stack full.");
  stream(sr)->ugstack[stream(sr)->ugstop] = c;
  stream(sr)->ugstop ++;
  stream(sr)->fill --;
  return 0;
  }

int cs_line_nb(ilstream sr)

  {
  return stream(sr)->line_nb;
  }

long srposition(ilstream sr)

  {
  if(streamid(sr)!=CHAR_FILE && streamid(sr)!=BINARY_FILE)
    return stream(sr)->fill;
  return ftell(stream(sr)->exte);
  }

void srseek(ilstream sr, long srp)

  {
  if(streamid(sr)!=CHAR_FILE && streamid(sr)!=BINARY_FILE)
    slex("srseek() called for a non-file or non-seekable file.");
  fseek(stream(sr)->exte, srp, SEEK_SET);
  }

cell slread(char *s)

  {
  cell ssr;
  cell rv;

  ssr = strconnect(MEMORY, s, INPUT);
  if(ssr==NIL)
    slex("Cannot connect input stream to memory for some reason.");
  rv = lread(ssr);
  disconnect(ssr);
  return rv;
  }

char *sprint(char *s, cell c, long sz)  /* returns new position in string */

  {
  cell ssr;
  char *rv;

  ssr = strconnect(MEMORY, s, OUTPUT);
  if(ssr==NIL)
    slex("Cannot connect input stream to memory for some reason.");
  dcr_mem_size(ssr, sz);
  print(c, ssr);
  rv = string_of(ssr)+stream(ssr)->fill;
  disconnect(ssr);
  return rv;
  }

char *sprin1(char *s, cell c, long sz)  /* returns new position in string */

  {
  cell ssr;
  char *rv;

  ssr = strconnect(MEMORY, s, OUTPUT);
  if(ssr==NIL)
    slex("Cannot connect input stream to memory for some reason.");
  dcr_mem_size(ssr, sz);
  set_p_pretty(0);
  prin1(c, ssr);
  rv = string_of(ssr)+stream(ssr)->fill;
  disconnect(ssr);
  return rv;
  }


char *spprint(char *s, cell c, long sz)  /* returns new position in string */

  {
  cell ssr;
  char *rv;

  ssr = strconnect(MEMORY, s, OUTPUT);
  if(ssr==NIL)
    slex("Cannot connect input stream to memory for some reason.");
  dcr_mem_size(ssr, sz);
  pprint(c, ssr);
  rv = string_of(ssr)+stream(ssr)->fill;
  disconnect(ssr);
  return rv;
  }


il5io()

  {
  csout = strconnect(CHAR_FILE, stdout, OUTPUT);
  csin = strconnect(CHAR_FILE, stdin, INPUT);
  cserr = strconnect(CHAR_FILE, stderr, OUTPUT);
  cskbd = strconnect(KEYBOARD, NULL, INPUT);
  }

/* End of Il5IO */


 
