/*    ILFIVE lisp objects library -- evaluator
 *
 *    Copyright (c) 1991--1998 Alexandru Dan Corlan M.D.
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    In short, you can use and modify this program. You can
 *    distribute the program with or without your modifications if you want,
 *    but you must not require licence fees and must provide the
 *    full source code with the distribution. If you make an application program
 *    which just calls functions from this library you may distribute that
 *    application program without source code, but also as an linkable object file, 
 *    and must distribute also this library with source code.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
\title{IL4LIFE - evaluator for ILFOUR}
\author{Alexandru Dan CORLAN}
\date{Jul 14, 1992}

\begin{abstract}
This is the first level of an evaluator of ILFOUR expressions.
Natures are implemented at the lowest level and a clean interface
is provided under which further extensions can be developed.
Compilation is not provided but place for it is left in the 
design.

The main (and final) design decisions are: provide natures,
make a code data type, use evaluation stack containing also symbol
bindings, a single environment binds symbols and primitives
and primitive function finding is done as in Scheme, implement
special forms in a special way, no lexical binding, no closures.
\end{abstract}

\section{Introduction}
This file happens to be the first written
using the new approach of simultaneus development of
program and documentation which was made possible by KNOST V5.0
(the FKnost tool). 

This module is the common denominator of the ILFIVE design
(made about an year ago) and the attempt to write a quick and
dirty evaluator that became IL4EV.

The approach is to make an evaluator aiming no less
than to the multitasking incremental compilation dream named
ILFIVE, but starting only with a simple core. This core
is to be easily mastered and implemented while further
development following to remain upward compatible with this.

\notion{il5life interface}

The functions that are exported by this module are:

    \hlink{push} --- a push a value on the stack.
    \hlink{drop} --- discard a value from the stack.
    \hlink{bind} --- push a value and a symbol on the stack.
    \hlink{sacc} --- symbol access --- return the address
                     of the symbol n places from the top.
    \hlink{vacc} --- value access --- same for value.
    \hlink{eval} --- given an \hlink{expression} return it's value in the
                     current context (and the single)
    \hlink{appl} --- given a \hlink{functional} apply it to the
                     values on the stack (their number also given).
    \hlink{makf} --- make a code function.
    \hlink{maks} --- make a [code] special form.
    \hlink{makm} --- make a code macro.
    
\notion{functional}

\subject{direct functional}
A direct functional is a representation of a function, special form
or macro. \subject{nature}
Natures are lists consisting of the symbol {\tt nature}
and two to four direct functionals.
Natures and direct functionals are called together functionals, 
i.e. are items that can be applied to lists of evaluated
or nonevaluated arguments. \subject{functional application}

Functionals fall in two cathegories: \subject{lambda form}
lambda forms (which are lists starting with the symbol
{\tt lambda}) and macros (starting with {\tt macro}) and, the
seccond cathegory \hlink[code object]{code objects},
which are ILFOUR elementary types of \hlink{id} CODE.
Special functions are always implemented as code functionals.

\notion{ILFOUR expression}
\subject{ILFOUR expression evaluation}
An ILFOUR expression is a any ILFOUR object which is given
to the eval function for evaluation. Objects other than 
non-keyword symbols and lists are, in this version,
their own value, i.e. are not changed during evaluation.

Symbols (which are not keywords --- they evaluate to themselves)
are looked for in the stack from top to bottom, and
the first binding found is returned as value.\footnote{
\subject{eval generalization} An interesting
extention would be permiting to place instead of symbol any
ILFOUR object; equal-ity between that object and the
expression given, if failing, would be followed by the usual 
evaluation procedure).

Lists are evaluated in the following way: the first element is evaluated.
\begin{itemize}
\item If the result is not a \hlink{functional}, then it is an error.

\item If it is a macro, the whole list is placed on the stack and
the macro functional is applied. It replaces the list with something else.
This is evaluated and the result is the result of the
evaluation. For the difference between code macros and 


\item If it is a special form, the whole list is placed on the stack
and the special form is called (unless it was treated in a
specific manner before). The special form always lefts one
item on the stack.

\item If it is a function, than all the other elements are evaluated
and placed on the stack so that the last comes to the top
of the stack. If the function is code it is called with the
number of elements on the stack. If not, a lambda expression interpreter
is called.

\item If it is a nature, then the full form is places on the stack.
The first nature method is called, being required to find what
functional can be used for evaluation of that construction. If one
is found, the construction is evaluated as if the first element would
have been that one (macro, special form, function or nature).
If none is found, then the elements are all evaluated and placed
on the stack. The secoond functional is called. If it returns
a functional it is applied (it has to be a function) to the elements.
Otherwise it is an error.

\end{itemize}


\section{The CODE type}
This piece of program implements code objects.

\notion{code object}

The code object is implemented a cell having in the first 
4 bytes a type-of-code (MACRO, SPECIAL_FORM or FUNCTION) and
in the seccond a pointer to that function (as a C function).

This function has one argument: 
number of items which are placed for it on the stack
when it is called. 

The function leaves ALWAYS one element
on the stack as a return value.

Some macros are defined in \hlink{shpell5.h}:

  \hlink{is_code} --- yields 1 if argument is a code object.
  \hlink{codetype} --- yields SPECIAL_FORM, FUNCTION or MACRO
                       given a code argument and 0 otherwise
  \hlink{codefun} --- yields a pointer to function returning
                      int, which is contained in the code object.
  \hlink{is_function} --- returns 1 if argument is code function.
  \hlink{is_special_form} --- returns 1 if argument is code special form.
  \hlink{is_macro} --- returns 1 if argument is macro.

\subsubsection{code constructors}

*/

#include <stdio.h>

#include <shpell5.h>
#include <il5vars.h>

cell seestack_;
cell just_;

cell makf(int (*f)(int n))  /* \subject{makf} */

  {
  cell rv;

  rv = lmake(CODE);
  codetype(rv) = FUNCTION;
  codefun(rv) = f;

  return rv;
  }

cell makm(int (*f)(int n))  /* \subject{makm} */

  {
  cell rv;

  rv = lmake(CODE);
  codetype(rv) = MACRO;
  codefun(rv) = f;

  return rv;
  }

cell maks(int (*f)(int n))  /* \subject{maks} */

  {
  cell rv;

  rv = lmake(CODE);
  codetype(rv) = SPECIAL_FORM;
  codefun(rv) = f;

  return rv;
  }


/*\section{the stack}

For the beginning there is only one single stack.
*/

#define STACKSIZE 512            /* \subject{STACKSIZE} */

static PSTEM stk[STACKSIZE];     /* the only stack \subject{stk} */
                                 /* see \hlink{CONTEXT} and \hlink{PSTEM} */

CONTEXT stop;  /* current top of the stack \subject{stop} */
static CONTEXT sbot;  /* bottom of the stack \subject{sbot} */
static CONTEXT slim;  /* limit of the stack \subject{slim} */

static cell slink_;   /* symbol for bottom of the stack \subject{slink_} */

/* \subsection{initialization} */

static initstack()    /* initialization of the stack, \subject{initstack} */

  {
  stop = sbot = stk;
  slim = stk + STACKSIZE - 1;
  slink_ = esym("bottom%of%the%stack");
  stop->stsym = stop->stval = slink_;
  stop ++;
  }

/* \subsection{specific functions} */

void push(cell v)    /* \subject{push} */

  {
  if(stop==slim)
    slex("Evaluation stack is full --- can't push anymore.");
  stop->stsym = NIL;
  stop->stval = v;
  stop ++;
  }

void drop()          /* \subject{drop} */

  {
  if(stop==sbot+1)
    slex("Evaluation stack is empty --- can't drop anymore.");
  stop --;
  }

void bind(cell s, cell v)   /* \subject{bind} */

  {
  if(stop==slim)
    slex("Evaluation stack is full --- can't bind anymore.");
  stop->stsym = s;
  stop->stval = v;
  stop ++;
  }

cell *sacc(int n)	                   /* subject{sacc} */

  {
  CONTEXT sp;

  if(stop-sbot-2<n)
    slex1i("Evaluation stack is not deep enough to sacc elem %d", n);
  sp = stop-n-1;  /* first is 0 */
  return &(sp->stsym);
  }

cell *vacc(int n)	                   /* subject{vacc} */

  {
  CONTEXT sp;

  if(stop-sbot-2<n)
    slex1i("Evaluation stack is not deep enough to vacc elem %d", n);
  sp = stop-n-1;  /* first is 0 */
  return &(sp->stval);
  }


cell pop()

  {
  stop --;
  if(stop==sbot)
    slex("Bottom of stack reached in pop().");
  return stop[0].stval;
  }

long binpop()

  {
  if(!is_binary(stop[-1].stval))
    slex("Binary required.");
  stop --;
  if(stop==sbot)
    slex("Bottom of stack reached in binpop().");
  return binary(stop[0].stval);
  }

/* \section{Evaluation} */

cell no_sym_val_;    /* returned if no value */

/* \subsection{symbol evaluation} */

cell symval(cell s)         /* \subject{symval} \subject{no_sym_val_} */
 
  {
  CONTEXT sp;

  for(sp=stop-1; sp>sbot; sp--)
    if(eq(sp->stsym,s))  
         /* change this to equal for \hlink{eval generalization} */
     	return sp->stval;
  return no_sym_val_;
  }

/* \subsection{the magic words} */

cell lambda_;       /* \subject{lambda_} */
cell macro_;        /* \subject{macro_} */
cell nature_;       /* \subject{nature_} */
cell setq;	        /* \subject{setq} */
cell quote;        /* \subject{quote} */

/* lambda interpretation must be written later */

int lambda_interpret(cell f, int n);  /* arguments are on the stack */
cell appl(cell f, int n);

/* \subsection{application of functionals} */

int lowappl(cell f, int n)  /* apply \hlink{functional} f to n elements */
                            /* \subject{lowappl} */

  {
  if(is_code(f))
    {
    if(is_special_form(f) || is_macro(f))
      if(n!=1) slex("lowapply on macros and special forms for 1 arg only.");
    return (* (codefun(f)))(n);
    }
  if(is_cons(f))
    {
    if(car(f)==lambda_)
      return lambda_interpret(f, n);     /* \subject{lambda_interpret}
                                            \hlink{lambda_interpret} */
    if(car(f)==macro_)
      if(n!=1)
        slex("macros can be applied only to exactly one arguments.");
      else
        return lambda_interpret(f, n);
   
    if(car(f)==nature_)
      {
      if(cdr(f)==NIL || cddr(f)==NIL)
        slex("Nature not containing apply method given to apply.");
      f = appl(caddr(f), n);  /* arguments remained on stack! */
      if(f==NIL)
        slex("Apply nature could not find a method in lowappl().");
      return lowappl(f, n);   
      }
    slex("if applying a list it must be macro, nature or lambda.");
    }
  }

/* \subject{appl} */

cell appl(cell f, int n)

  {
  cell rv;

  n = lowappl(f, n);
  if(n!=1)
    slex("C interface only works with functions returning exactly 1 result");
  rv = *(vacc(0));
  drop();
  return rv;
  }

/* \subject{map} --- very similar to apply, but a Lisp primitive
 *                   that can't (and needn't) be used from C 
 *                   for the moment does not work with vectors,
 *                   but only with lists */

int themap(int n)

  {
  long l;
  int j;
  int newn;
  cell c;
  cell f;
  int i;
  CONTEXT lsts;
  CONTEXT sp;

  if(n==-1)
    slex("map can only be used in Lisp mode.");  /* i.e. no FORTH */
  f = *vacc(n-1);
  l = -1;  /* not yet defined */
  newn = 0;
  for(i=n-2; i>=0; i--) /* checking the size of the given lists */
    {
    if((c = *vacc(i))==just_)
      {
      i--;
      newn++;
      continue;
      }
    if(!is_list(c))
      slex("In this version map works only with lists.");
    if(l==-1)
      l = length(c);
    else
      if(l!=length(c))
        slex("Lists given to map do not have the same length.");
    newn++;
    }
  /* correct arguments --- we can start processing */
  if(l==-1)
    slex("map doesn't make sense without lists to map onto.");
  lsts = stop;
  for(j=0; j<l; j++)
    {
    for(i=n-2; i>=0; i--)
      {
      sp = lsts-i-1;
      if(sp->stval==just_)
        {
        i--;
        sp = lsts-i-1;
        push(sp->stval);
        continue;
        }
	  c = car(sp->stval);
     sp->stval = cdr(sp->stval);
     push(c);
     }
    appl(f, newn);
    /* the result is ignored in map */
    }
  for(i=0; i<n; i++)
    drop();
  push(NIL);
  return 1;
  }

/* \subject{mapcar}  very similar to apply, but a Lisp primitive
 *                   that can't (and needn't) be used from C 
 *                   for the moment does not work with vectors,
 *                   but only with lists */

int themapcar(int n)

  {
  long l;
  int j;
  int newn;
  cell c;
  cell f;
  int i;
  CONTEXT lsts;
  CONTEXT sp;

  if(n==-1)
    slex("mapcar can only be used in Lisp mode.");  /* i.e. no FORTH */
  f = *vacc(n-1);
  l = -1;  /* not yet defined */
  newn = 0;
  for(i=n-2; i>=0; i--) /* checking the size of the given lists */
    {
    if((c = *vacc(i))==just_)
      {
      i--;
      newn++;
      continue;
      }
    if(!is_list(c))
      slex("In this version mapcar works only with lists.");
    if(l==-1)
      l = length(c);
    else
      if(l!=length(c))
        slex("Lists given to map do not have the same length.");
    newn++;
    }
  /* correct arguments --- we can start processing */
  if(l==-1)
    slex("mapcar doesn't make sense without lists to map onto.");
  lsts = stop;
  for(j=0; j<l; j++)
    {
    for(i=n-2; i>=0; i--)
      {
      sp = lsts-i-1;
      if(sp->stval==just_)
        {
        i--;
        sp = lsts-i-1;
        push(sp->stval);
        continue;
        }
      c = car(sp->stval);
      sp->stval = cdr(sp->stval);
      push(c);
      }
    i = lowappl(f, newn);
    if(i!=1)
      slex("Can only mapcar functions returning exactly one result.");
    }
  c = NIL;
  for(i=0; i<l; i++)
    c = cons(pop(), c);
  for(i=0; i<n; i++)
    drop();
  push(c);
  return 1;
  }



/* \notion{leval} */

cell leval(cell expr);

static int eval_and_push(cell l)

  {
  int r;

  for(r=0; l!=NIL; r++, l=cdr(l))
    push(leval(car(l)));
  return r;
  }

cell leval(cell expr)

  {
  cell rv;
  cell m;
  CONTEXT sp;

  if(is_keyword(expr)) 
    return expr;

  if(is_symbol(expr))
    if((rv = symval(expr))==no_sym_val_)
      slex1s("Symbol %s not bound in leval().", symname(expr));
    else
		return rv;

  if(!is_cons(expr))
    return expr;      /* THIS IS GOING TO CHANGE GENTLEMEN! */
                      /* \subject{warning} \subject{gentlemen} */
                      /* \subject{eval generalization} */

/* Some special forms are treated here for efficiency and
 * more clarity of the code; it is important that they can't
 * be redefined in any way
 */

  if(car(expr)==seestack_)
    {
    for(sp=stop; sp>sbot; sp--)
    	{
      print(sp->stsym, cserr);
      pprint(sp->stval, cserr);
      }
    return NIL;
    }

  if(car(expr)==quote)
    return cadr(expr);

  if(car(expr)==setq)
    {
    m=NIL;
    for(expr=cdr(expr); expr!=NIL; expr=cddr(expr))
      {
      if(!is_cons(cdr(expr)))
        slex("setq arguments must come in pairs.");
      for(sp=stop; sp>sbot; sp--)
        if(car(expr)==sp->stsym)
          {
          m = (sp->stval = leval(cadr(expr)));
          break;
          }
      if(sp==sbot)
        bind(car(expr), m = leval(cadr(expr)));    
      }
    return m;
    }
  
  rv = leval(car(expr));

  if(is_code(rv))
    {
    if(is_macro(rv))                    /* \subject{macro} */
      {
      push(expr);
      rv = appl(rv, 1);
      return leval(rv);
      }
    if(is_special_form(rv))				  /* \subject{special form} */
      {
      push(expr);
      return appl(rv, 1);
      }
    }
  if(is_cons(rv))     /* it must be a lambda or a macro */
    {
    if(car(rv)==macro_)  /* \subject{macro_} */
      {
      push(expr);
      rv = appl(rv, 1);
      return leval(rv);
      }
    if(car(rv)==nature_)
      {
		if(!is_cons(cdr(rv)))
        slex("Nature does not have an eval method");
      if(cadr(rv)!=NIL)
        { /* try the eval nature */
        push(expr);
		  m = appl(cadr(rv), 1);
        if(m!=NIL)  /* this is the method to use */
          return appl(m, eval_and_push(cdr(expr)));
        }
      if(!is_cons(cddr(rv)))
        slex("Nature does not have an apply method and eval method didn't work.");
      return appl(rv, eval_and_push(cdr(expr)));
      }
    return appl(rv, eval_and_push(cdr(expr)));
    }
  return appl(rv, eval_and_push(cdr(expr)));
  }

/* \subsection{the lambda form interpreter} */

static void lambdalist(cell l, int n)     /* \subject{lambdalist} */

/*  This function binds symbols from the lambda list to
 *  the objects that are placed on the stack; it raises
 *  an exception if the lambda form is badly constructed.
 */

  {
  int i;

  /* In this version only fix lists are allowed */
  for(i=1; i<=n && l!=NIL; l=cdr(l), i++)
    *sacc(n-i) = car(l);
  if(l!=NIL)
    slex("Too few arguments.");
  if(i<=n)
    slex("Too many arguments.");
  }			 

int lambda_interpret(cell f, int n)  /* \subject{lambda_interpret} */

  {
  cell rv;

  lambdalist(cadr(f), n);
  for(f=cddr(f); f!=NIL; f = cdr(f))
    rv = leval(car(f));  /* assumed to have left the stack untouched */
  for(;n>0;n--) drop();
  push(rv);
  return 1;      /* a \hlink{values} special form would be treated here
                  * in a particular way for the return of many values
                  */
  }

/* \section{Interface to eval} */

int theeval(int n)

  {
  if (n==-1)
    n = 1;
  if(n!=1)
    slex("eval expects exactly one argument.");
  push(leval(pop()));
  return 1;
  }

/* \section{initialization of il5life} */

il5life()

  {
  initstack();											/* \hlink{initstack} */
  seestack_ = esym("seestack");
  just_ = esym(":just");
  no_sym_val_ = esym("no%%symbol%%value");   /* \subject{no_sym_val_} */
  lambda_ = esym("lambda");                  /* \subject{lambda_} */
  macro_ = esym("macro");						   /* \subject{macro_}*/
  nature_ = esym("nature");                  /* \subject{nature_} */
  setq = esym("setq");	                     /* \subject{setq} */
  quote = esym("quote");                     /* \subject{quote} */

  bind(esym("map"), makf(themap));
  bind(esym("mapcar"), makf(themapcar));
  bind(esym("eval"), makf(theeval));
  }

/* End of the il5life interpreter */


 
