/*    ILFIVE lisp objects library -- minimal interpreter primitives
 *
 *    Copyright (c) 1991--1998 Alexandru Dan Corlan M.D.
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    In short, you can use and modify this program. You can
 *    distribute the program with or without your modifications if you want,
 *    but you must not require licence fees and must provide the
 *    full source code with the distribution. If you make an application program
 *    which just calls functions from this library you may distribute that
 *    application program without source code, but also as an linkable object file, 
 *    and must distribute also this library with source code.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
\title{IL4MINIM - minimal functions for il5life}
\author{Alexandru Dan CORLAN}
\date{Jul 14, 1992}

\begin{abstract}
This module contains a host of functions that offer
minimal arithmetical, logic, type coercion and 
data comparing, and also list manipulation. 

Each of these sections of the file are extended 
with more functions in other files. 

All functions work only with the most basic datatypes; functions
for complex numbers are not included (but for the arithmetical
operations should be added here).
\end{abstract}

\section{Overview}

\subject{il5minim}

This module contains a host of functions that offer
minimal arithmetical, logic, type coercion and data comparing,
and also list manipulation. Functions are designed such
that the smallest possible number of them are used; for example
{\tt <} is the same function for integers and strings. Automatic
data coercion is provided when possible; for example one 
can compare strings and symbols --- their names are compared
for Ascii alphabetical order.

The file il5minim.til, if run, must give the same result as
il5minim.out --- this being the test file. It is also a good and
fast documentation of how these functions are to used.

The functions are:

    The four arithmetic operations \ref{fao}:
{\tt

    +           \hlink{plus}
    -           \hlink{minus}
    *           \hlink{multiply}
    /           \hlink{divide}
}

\subject{extention}

    These functions are to be completed in the following way:
\begin{itemize}

\item numeric \hlink{vector} and \hlink{complex number}
 equivalents for the four
arithmetic operations, and also such functions as modulus
are to be added to this file.

\item \hlink{trigonometric function}s, square root, exponentials and
the like are to be added in a package named \hlink{il5math}, for
all fundamental data types (binary, float, complex).

\item \hlink{scatter} and \hlink{gather}, functions on polynomes
and other vectorial primitives
are to be kept in a module named \hlink{il5apl}.

\item symbolic calculus functions should be kept in \hlink{il5deriv}.

\item output of formulas as \TeX expressions should be kept
in \hlink{il5tex}.

\end{itemize}

    The six logical comparisons \ref{slc}:

{\tt
    {\tt <}     \hlink{lt}
    {\tt >}     \hlink{gt}
    {\tt <=}    \hlink{le}
    {\tt >=}    \hlink{ge}
    {\tt ==}    \hlink{ee}
    {\tt !=}    \hlink{ne}
}

   The three fundamental logical functions \ref{flf}:

{\tt
    \verb|&&|   \hlink{andand}
    \verb/||/   \hlink{oror}
    \verb|!|    \hlink{notnot}
}

   The basic functions for list processing \ref{bflp}:

{\tt
	 \verb|car|     \hlink{thecar}    
    \verb|cdr| 	 \hlink{thecdr}	 \verb|eq|        \hlink{theeq}
    \verb|cons|	 \hlink{thecons}	 \verb|eql|       \hlink{theeql}
    \verb|list|	 \hlink{thelist}	 \verb|equal|     \hlink{theequal}
    \verb|append|  \hlink{theappend} \verb|copy|      \hlink{thecopy}
}

    Elementary \hlink{type coercion} function \ref{etc}:
	 \subject{type coercion}

{\tt \def\*{$\bullet$}
\begin{tabular}{l|c|c|c|c|c|c}
\hline
$\downarrow$\ from, to $\rightarrow$ &
       bin & sym & str & char & list & vect \\
\hline
bin  &  -  &  \* & \*  &  \*  &      &      \\
\hline
sym  &     &  -  & \*  &  \*  &  \*  & \*   \\
\hline
str  &  \* &  \* & -   &  \*  &  \*  & \*   \\
\hline
char &  \* &  \* & \*  &  -   &      & 	  \\
\hline
list &     &  \* & \*  &      &  -   & \*   \\
\hline
vect &     &  \* & \*  &      &  \*  &  -   \\
\hline
\end{tabular}
}

It is the function \verb|coerce| 

\section{External declarations}

*/

#include <stdio.h>

#include <shpell5.h>

/*
\section{The four arithmetical operations}\label{fao}

These functions have the classical $-$, $+$, $*$ and $/$
names. They work for both \hlink{binary} and \hlink{float}
data type, which can be mixed in any order.

\notion{plus}

This function can take any number of arguments. If none, 
0 binary is returned; if one, that argument is returned;
if more, their algebric sum is returned; if one or more
are \hlink{float}, a float is returned; 
if none is float, a \hlink{binary} is returned.

*/

int plus(int n)

  {
  double r;
  cell o;
  int i;
  int flomode;
  
  r = 0.0;
  flomode = 0;
  if(n==-1) 
    n = 2;
  for(i=0; i<n; i++)
	 {
    o = *vacc(0);
    if(!is_binary(o) && !is_float(o))
      slex("+ works only with numeric arguments.");
    if(is_float(o))
      {
      flomode = 1;
      r += value(o);
      }
    else
      r += binary(o);
    drop();
    }
  if(flomode)
    {
    o = lmake(FLOAT);
	 value(o) = r;
    push(o);
    }
  else
    push(make(BINARY, (long)r));
  return 1;
  }

/*
\notion{multiply}

This function can take any number of arguments. If none, 
1 binary is returned; if one, that argument is returned;
if more, their product is returned; if one or more
are \hlink{float}, a float is returned; 
if none is float, a \hlink{binary} is returned.
*/

int multiply(int n)

  {
  double r;
  cell o;
  int i;
  int flomode;
  
  r = 1.0;
  flomode = 0;
  if(n==-1) 
    n = 2;
  for(i=0; i<n; i++)
	 {
    o = *vacc(0);
    if(!is_binary(o) && !is_float(o))
      slex("* works only with numeric arguments.");
    if(is_float(o))
      {
      flomode = 1;
      r *= value(o);
      }
    else
      r *= binary(o);
    drop();
    }
  if(flomode)
    {
    o = lmake(FLOAT);
	 value(o) = r;
    push(o);
    }
  else
    push(make(BINARY, (long)r));
  return 1;
  }

/*
\notion{minus}

This function can take one or more arguments. If none,
it is an error; if one, that argument is returned with
it;s sign changed;
if more, the result of substracting all that follow
from the first one is returned; 
if one or more are \hlink{float}, a float is returned; 
if none is float, a \hlink{binary} is returned.
*/


int minus(int n)

  {
  double r;
  cell o;
  int i;
  int flomode;
  
  flomode = 0;
  if(n==-1) 
    n = 2;
  if (n<1)
    slex("- needs at least one argument.");
  for(i=n-1; i>=0; i--)
	 {
    o = *vacc(i);
    if(!is_binary(o) && !is_float(o))
      slex("- works only with numeric arguments.");
    if(is_float(o))
      {
      flomode = 1;
      if(i<n-1)
        r -= value(o);
      else
		  r = value(o);
      }
    else
      if(i<n-1)
        r -= binary(o);
      else
        r = binary(o);
    }
  if(n==1)
    r = -r;
  for(i=0; i<n; i++)
    drop();
  if(flomode)
    {
    o = lmake(FLOAT);
	 value(o) = r;
    push(o);
    }
  else
    push(make(BINARY, (long)r));
  return 1;
  }

/*
\notion{divide}

This function can take two or more arguments. If none, or 
only one it is an error; 
if more, the result of dividing the first to all that follow
is returned; if one or more are \hlink{float}, a float is returned; 
if none is float, a \hlink{binary} is returned, but anyway all
intermediary computations are done in float.
*/


int divide(int n)

  {
  double r;
  cell o;
  int i;
  int flomode;
  
  flomode = 0;
  if(n==-1) 
    n = 2;
  if (n<2)
    slex("/ needs at least two arguments.");
  for(i=n-1; i>=0; i--)
	 {
    o = *vacc(i);
    if(!is_binary(o) && !is_float(o))
      slex("/ works only with numeric arguments.");
    if(is_float(o))
      {
      flomode = 1;
      if(i<n-1)
        r /= value(o);
      else
		  r = value(o);
      }
    else
      {
      if(i<n-1)
        r /= binary(o);
      else
        r = binary(o);
      }
    }
  for(i=0; i<n; i++)
    drop();
  if(flomode)
    {
    o = lmake(FLOAT);
	 value(o) = r;
    push(o);
    }
  else
    push(make(BINARY, (long)r));
  return 1;
  }

/*
\section{The six logical comparisons}\label{slc}

These functions compare between them similar item, as for example
symbols and strings or numbers. Equality of numbers is equality of
their value; equality of \hlink{string}s is identity in number and contents
of the characters that make them up; \hlink{symbol}s are treated in this
respect just like strings, their \hlink[name]{symbol name}
(irrespective of \hlink{package}) is taken into account.
Symbols and strings can be mixed freely, and also can binaries
and floats.

All these functions return either the symbol {\tt t} or nil,
the void list.

\notion{ge}

\subject{sort}

This function returns {\tt t} if arguments are in not-strict
descending order or if they are fewer than 2, and NIL otherwise.

*/

int ge(int n)

  {
  double r;
  cell o;
  int i;
  int rv;
  char *sr;
  int typ;
  
  
  if(n==-1)
    n = 2;
  if(n==0 || n==1)
    {
    for(;n>0;n--) drop();
    push(t_);
    return 1;
    }
  o = *vacc(0);
  typ = id(o);
  switch(typ)
    {
    case STRING:
      sr = string(o);
      break;
    case SYMBOL:
      sr = symname(o);
      break;
    case BINARY:
      r = binary(o);
      break;
    case FLOAT:
      r = value(o);
      break;
    default:
      slex(">= works only with numeric, string or symbol arguments.");
    }
  rv = 1;
  drop(); n--;
  for(i=0; i<n; i++)
    {
    o = *vacc(0);
    switch(id(o))
      {
      case BINARY:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r <= binary(o))
          r = binary(o);
        else
          rv = 0;
        break;
      case FLOAT:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r <= value(o))
          r = value(o);
        else
          rv = 0;
        break;
      case STRING:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr,string(o))<=0)
          sr = string(o);
        else
          rv = 0;
        break;
      case SYMBOL:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr, symname(o))<=0)
          sr = symname(o);
        else
          rv = 0;
        break;
      default:
        slex(">= works only with numeric, string or symbol arguments.");
      }
    drop();
    }
  if(rv)
    push(t_);
  else
    push(NIL);
  return 1;
badtypes:
  slex(">= arguments given can't be compared.");
  }

/*
\notion{gt}

\subject{sort}

This function returns {\tt t} if arguments are in strict
descending order or if they are fewer than 2, and NIL otherwise.

*/


int gt(int n)

  {
  double r;
  cell o;
  int i;
  int rv;
  char *sr;
  int typ;
  
  
  if(n==-1)
    n = 2;
  if(n==0 || n==1)
    {
    for(;n>0;n--) drop();
    push(t_);
    return 1;
    }
  o = *vacc(0);
  typ = id(o);
  switch(typ)
    {
    case STRING:
      sr = string(o);
      break;
    case SYMBOL:
      sr = symname(o);
      break;
    case BINARY:
      r = binary(o);
      break;
    case FLOAT:
      r = value(o);
      break;
    default:
      slex("> works only with numeric, string or symbol arguments.");
    }
  rv = 1;
  drop(); n--;
  for(i=0; i<n; i++)
    {
    o = *vacc(0);
    switch(id(o))
      {
      case BINARY:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r < binary(o))
          r = binary(o);
        else
          rv = 0;
        break;
      case FLOAT:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r < value(o))
          r = value(o);
        else
          rv = 0;
        break;
      case STRING:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr,string(o))<0)
          sr = string(o);
        else
          rv = 0;
        break;
      case SYMBOL:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr, symname(o))<0)
          sr = symname(o);
        else
          rv = 0;
        break;
      default:
        slex("> works only with numeric, string or symbol arguments.");
      }
    drop();
    }
  if(rv)
    push(t_);
  else
    push(NIL);
  return 1;
badtypes:
  slex("> arguments given can't be compared.");
  }

/*
\notion{le}

\subject{sort}

This function returns {\tt t} if arguments are in non-strict
ascending order or if they are fewer than 2, and NIL otherwise.

*/


int le(int n)

  {
  double r;
  cell o;
  int i;
  int rv;
  char *sr;
  int typ;
  
  
  if(n==-1)
    n = 2;
  if(n==0 || n==1)
    {
    for(;n>0;n--) drop();
    push(t_);
    return 1;
    }
  o = *vacc(0);
  typ = id(o);
  switch(typ)
    {
    case STRING:
      sr = string(o);
      break;
    case SYMBOL:
      sr = symname(o);
      break;
    case BINARY:
      r = binary(o);
      break;
    case FLOAT:
      r = value(o);
      break;
    default:
      slex("<= works only with numeric, string or symbol arguments.");
    }
  rv = 1;
  drop(); n--;
  for(i=0; i<n; i++)
    {
    o = *vacc(0);
    switch(id(o))
      {
      case BINARY:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r >= binary(o))
          r = binary(o);
        else
          rv = 0;
        break;
      case FLOAT:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r >= value(o))
          r = value(o);
        else
          rv = 0;
        break;
      case STRING:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr,string(o))>=0)
          sr = string(o);
        else
          rv = 0;
        break;
      case SYMBOL:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr, symname(o))>=0)
          sr = symname(o);
        else
          rv = 0;
        break;
      default:
        slex("<= works only with numeric, string or symbol arguments.");
      }
    drop();
    }
  if(rv)
    push(t_);
  else
    push(NIL);
  return 1;
badtypes:
  slex("<= arguments given can't be compared.");
  }

/*
\notion{lt}

\subject{sort}

This function returns {\tt t} if arguments are in strict
ascending order or if they are fewer than 2, and NIL otherwise.

*/


int lt(int n)

  {
  double r;
  cell o;
  int i;
  int rv;
  char *sr;
  int typ;
  
  
  if(n==-1)
    n = 2;
  if(n==0 || n==1)
    {
    for(;n>0;n--) drop();
    push(t_);
    return 1;
    }
  o = *vacc(0);
  typ = id(o);
  switch(typ)
    {
    case STRING:
      sr = string(o);
      break;
    case SYMBOL:
      sr = symname(o);
      break;
    case BINARY:
      r = binary(o);
      break;
    case FLOAT:
      r = value(o);
      break;
    default:
      slex("< works only with numeric, string or symbol arguments.");
    }
  rv = 1;
  drop(); n--;
  for(i=0; i<n; i++)
    {
    o = *vacc(0);
    switch(id(o))
      {
      case BINARY:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r > binary(o))
          r = binary(o);
        else
          rv = 0;
        break;
      case FLOAT:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r > value(o))
          r = value(o);
        else
          rv = 0;
        break;
      case STRING:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr,string(o))>0)
          sr = string(o);
        else
          rv = 0;
        break;
      case SYMBOL:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr, symname(o))>0)
          sr = symname(o);
        else
          rv = 0;
        break;
      default:
        slex("< works only with numeric, string or symbol arguments.");
      }
    drop();
    }
  if(rv)
    push(t_);
  else
    push(NIL);
  return 1;
badtypes:
  slex("< arguments given can't be compared.");
  }

/*
\notion{ee}

\subject{equality}

This function returns {\tt t} if arguments are all
equal in value 
or if they are fewer than 2, and NIL otherwise.

*/


int ee(int n)

  {
  double r;
  cell o;
  int i;
  int rv;
  char *sr;
  int typ;
  
  
  if(n==-1)
    n = 2;
  if(n==0 || n==1)
    {
    for(;n>0;n--) drop();
    push(t_);
    return 1;
    }
  o = *vacc(0);
  typ = id(o);
  switch(typ)
    {
    case STRING:
      sr = string(o);
      break;
    case SYMBOL:
      sr = symname(o);
      break;
    case BINARY:
      r = binary(o);
      break;
    case FLOAT:
      r = value(o);
      break;
    default:
      slex("== works only with numeric, string or symbol arguments.");
    }
  rv = 1;
  drop(); n--;
  for(i=0; i<n; i++)
    {
    o = *vacc(0);
    switch(id(o))
      {
      case BINARY:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r == binary(o))
          r = binary(o);
        else
          rv = 0;
        break;
      case FLOAT:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r == value(o))
          r = value(o);
        else
          rv = 0;
        break;
      case STRING:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr,string(o))==0)
          sr = string(o);
        else
          rv = 0;
        break;
      case SYMBOL:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr, symname(o))==0)
          sr = symname(o);
        else
          rv = 0;
        break;
      default:
        slex("== works only with numeric, string or symbol arguments.");
      }
    drop();
    }
  if(rv)
    push(t_);
  else
    push(NIL);
  return 1;
badtypes:
  slex("== arguments given can't be compared.");
  }

/*
\notion{ne}

\subject{equality}

This function returns {\tt t} if arguments are 
not equal in value --- i.e. if there is any pair of
different values
or if they are fewer than 2, and NIL otherwise.

*/


int ne(int n)

  {
  double r;
  cell o;
  int i;
  int rv;
  char *sr;
  int typ;
  
  
  if(n==-1)
    n = 2;
  if(n!=2)
    slex("!= must have exactly two arguments.");
  o = *vacc(0);
  typ = id(o);
  switch(typ)
    {
    case STRING:
      sr = string(o);
      break;
    case SYMBOL:
      sr = symname(o);
      break;
    case BINARY:
      r = binary(o);
      break;
    case FLOAT:
      r = value(o);
      break;
    default:
      slex("!= works only with numeric, string or symbol arguments.");
    }
  rv = 1;
  drop(); n--;
  for(i=0; i<n; i++)
    {
    o = *vacc(0);
    switch(id(o))
      {
      case BINARY:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r != binary(o))
          r = binary(o);
        else
          rv = 0;
        break;
      case FLOAT:
        if(typ!=BINARY && typ!=FLOAT)
          goto badtypes;
        if(r != value(o))
          r = value(o);
        else
          rv = 0;
        break;
      case STRING:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr,string(o))!=0)
          sr = string(o);
        else
          rv = 0;
        break;
      case SYMBOL:
        if(typ!=STRING && typ!=SYMBOL)
          goto badtypes;
        if(strcmp(sr, symname(o))!=0)
          sr = symname(o);
        else
          rv = 0;
        break;
      default:
        slex("!= works only with numeric, string or symbol arguments.");
      }
    drop();
    }
  if(rv)
    push(t_);
  else
    push(NIL);
  return 1;
badtypes:
  slex("!= arguments given can't be compared.");
  }


/*

\section{The fundamental logical functions}\label{flf}

These are the three fundamental logical functions, applyable
to any number of arguments (except of course the \hlink{not} function).
They are bound to symbols similar to the C operator names.

\subject{logical value}
The arguments are considered to be logically true if they are
not nil, and are considered false if they are nil.
The result is the {\tt t} symbol for a true result and {\tt ()}
for a false one.

\notion{andand}

This function returns {\tt t} if all the arguments are not-nil
and nil if at least one argument is nil.

*/

int andand(int n)

  {
  int i;
  int rv;

  rv = 1;
  if(n==-1)
    n = 2;
  for(i=0; i<n; i++, drop())
    if(*vacc(0)==NIL)
      rv = 0;
  if(rv)
    push(t_);
  else
    push(NIL);
  return 1;
  }

/*

\notion{oror}

This function returns {\tt ()} if all the arguments are nil
and {\tt t} if at least one argument is not nil.

*/

int oror(int n)

  {
  int i;
  int rv;

  rv = 0;
  if(n==-1)
    n = 2;
  for(i=0; i<n; i++, drop())
    if(*vacc(0)!=NIL)
      rv = 1;
  if(rv)
    push(t_);
  else
    push(NIL);
  return 1;
  }

/*

\notion{notnot}

This function returns {\tt ()} if all the only argument
is {\tt t} and returns {\tt t} otherwise.

*/

int notnot(int n)

  {
  if(n==-1)
    n = 1;
  if(n!=1)
    slex("! (not) can only take one argument.");
  if(*vacc(0)==NIL)
    *vacc(0) = t_;
  else
    *vacc(0) = NIL;
  *sacc(0) = NIL;
  return 1;
  }

/*
\section{basic functions for list processing}\label{bflp}

These are the most fundamental of the traditional Lisp functions.
Others can be found in more elaborate (optionally loaded)
packages.

*/

/* \notion{thecar} \subject{car} */

int thecar(int n)

  {
  cell c;

  if(n==-1) n=1;
  if(n!=1)
    slex("CAR takes just one argument.");
  c = *vacc(0);
  if(!is_cons(c))
    slex("CAR takes only a list argument.");
  *vacc(0) = car(c);
  *sacc(0) = NIL;
  return 1;
  }

/* \notion{thecdr} \subject{cdr} */

int thecdr(int n)

  {
  cell c;

  if(n==-1) n=1;
  if(n!=1)
    slex("CDR takes just one argument.");
  c = *vacc(0);
  if(!is_cons(c))
    slex("CDR takes only a list argument.");
  *vacc(0) = cdr(c);
  *sacc(0) = NIL;
  return 1;
  }

/* \notion{thecons} \subject{cons} */

int thecons(int n)

  {
  if(n==-1) n=1;
  if(n!=2)
    slex("CONS takes exactly two arguments.");
  *vacc(1)=cons(*vacc(1), *vacc(0));
  *sacc(1) = NIL;
  drop();
  return 1;
  }

/* \notion{thelist} \subject{list} 

This function takes zero or more arguments, which are transformed
into a list in the order of their occourence in the call.

*/

int thelist(int n)

  {
  cell rv;
  int i;

  if(n==-1)
    slex("LIST cannot be called as a FORTH function.");
  for(rv=NIL, i=0; i<n; i++)
    {
    rv = cons(*vacc(0), rv);
    drop();
    }
  push(rv);
  return 1;
  }

/* \notion{theappend}\subject{append}

This function takes a (possibly empty) list of lists and
transforms them into a list that is the concatenation of all
of them.

*/

static cell lappend(cell h, cell t)

  {
  if(h==NIL)
    return t;
  return cons(car(h), lappend(cdr(h), t));
  }

int theappend(int n)

  {
  int i;
  cell rv;

  rv = NIL;
  if(n==-1)
    slex("APPEND cannot be just called in FORTH mode.");
  for(i=0; i<n; i++)
    {
    if(!is_list(*vacc(0)))
      slex("We can only append lists.");
    rv = lappend(*vacc(0), rv);
    drop();
    }
  push(rv);
  return 1;
  }

/* \notion{theeq}\subject{eq} 

This function takes two arguments and returns {\tt t} if
they are the same cell and NIL otherwise.

*/

int theeq(int n)

  {
  if(n==-1)
    n=2;
  if(n!=2)
    slex("EQ can take only exactly 2 arguments.");
  if(*vacc(0)==*vacc(1))
    *vacc(1) = t_;
  else
    *vacc(1) = NIL;
  drop();
  return 1;
  }

/* \notion{theeql}\subject{eql} 

This function takes two arguments and returns {\tt t} if
they are eql (see \blink{Guy;Steele;Common;Lang;first})
and NIL otherwise.

*/

int theeql(int n)

  {
  if(n==-1)
    n=2;
  if(n!=2)
    slex("EQL can take only exactly 2 arguments.");
  if(eql(*vacc(0),*vacc(1)))
    *vacc(1) = t_;
  else
    *vacc(1) = NIL;
  drop();
  return 1;
  }

/* \notion{theequal}\subject{equal} 

This function takes two arguments and returns {\tt t} if
they are eql (see \blink{Guy;Steele;Common;Lang;first})
and NIL otherwise.

*/

int theequal(int n)

  {
  if(n==-1)
    n=2;
  if(n!=2)
    slex("EQUAL can take only exactly 2 arguments.");
  if(equal(*vacc(0),*vacc(1)))
    *vacc(1) = t_;
  else
    *vacc(1) = NIL;
  drop();
  return 1;
  }

/* \notion{thecopy}\subject{copy}

This function takes one argument which it replaces
by an identical copy with new cells
 
*/

int thecopy(int n)

  {
  if(n==-1)
    n=1;
  if(n!=1)
    slex("COPY can copy one object only.");
  *vacc(0) = copy(*vacc(0));
  *sacc(0) = NIL;
  return 1;
  }

/* \section{Type coercion}

The only function coerce  \subject{coerce} is given a type, then
a value of another type and returns a value ``with the same implementation''
as the first type.

*/

static cell Binary_;
static cell Char_;
static cell Float_;
static cell List_;

cell coerce(int typ, cell v)

  {  /* This function can theoretically be extended for user types */
  cell rv;
  long l;
  int i;

  switch(typ)
    {
    case VECTOR:
      switch(id(v))
        {
        case LIST:
          rv = make_vector(l = length(v));
          for(i=0; v!=NIL && i<l; i++, v = cdr(v))
            vnth(rv,i) = car(v);
          return rv;
        default:
          slex1i("Can't convert objects of type %d to VECTOR.", id(v));
        }
    case BINARY:
      switch(id(v))
        {
        case CHAR:
          return make(BINARY, v->charcell.ccode);
        default:
          slex1i("Can't convert objects of type %d to BINARY.", id(v));
        }
    case CHAR:
      switch(id(v))
        {
        case BINARY:
          rv = lmake(CHAR);
          rv->charcell.ccode = (int)binary(v);
          return rv;
        default:
          slex1i("Can't convert objects of type %d to CHAR.", id(v));
        }
    default:
      slex1i("Can't convert to type %d.", typ);
    }
  }

int thecoerce(int n)

  {
  cell v;
  cell ts;
  int t;

  v = *vacc(0);
  drop();
  ts = *vacc(0);
  if(ts==Binary_)
    { 
    t = BINARY;
    goto thec1;
    }
  if(ts==Char_)
    {
    t = CHAR;
    goto thec1;
    }
  if(ts==Float_)
    {
    t = FLOAT;
    goto thec1;
    }
  if(ts==List_)
    {
    t = LIST;
    goto thec1;
    }
  slex("Target coercion type is unknown.");
thec1:
  drop();
  push(coerce(t, v));
  return 1;
  }

il5minim()

  {
  t_ = esym("t");

  /* THE NAMES OF THE TYPES */

  Binary_ = esym("binary");
  Char_ = esym("char");
  Float_ = esym("float");
  List_ = esym("list");

  bind(esym("/"), makf(divide));
  bind(esym("-"), makf(minus));
  bind(esym("*"), makf(multiply));
  bind(esym("+"), makf(plus));

  bind(esym("<"), makf(lt));
  bind(esym(">"), makf(gt));
  bind(esym("<="), makf(le));
  bind(esym("=="), makf(ee));
  bind(esym("!="), makf(ne));
  bind(esym(">="), makf(ge));

  bind(esym("&&"), makf(andand));
  bind(esym("||"), makf(oror));
  bind(esym("!"), makf(notnot));

  bind(esym("car"), makf(thecar));
  bind(esym("cdr"), makf(thecdr));
  bind(esym("cons"), makf(thecons));
  bind(esym("list"), makf(thelist));
  bind(esym("append"), makf(theappend));
  bind(esym("eq"), makf(theeq));
  bind(esym("eql"), makf(theeql));
  bind(esym("equal"), makf(theequal));
  bind(esym("copy"), makf(thecopy));

  bind(esym("coerce"), makf(thecoerce));
  }

