/*    ILFIVE lisp objects library -- lists as sets
 *
 *    Copyright (c) 1991--1998 Alexandru Dan Corlan M.D.
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    In short, you can use and modify this program. You can
 *    distribute the program with or without your modifications if you want,
 *    but you must not require licence fees and must provide the
 *    full source code with the distribution. If you make an application program
 *    which just calls functions from this library you may distribute that
 *    application program without source code, but also as an linkable object file, 
 *    and must distribute also this library with source code.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>

#include <shpell5.h>

int memq(cell obj, cell set)

  {
  while(set!=NIL)
    {
    if (eq(obj,car(set)))
      return 1;
    set = cdr(set);
    }
  return 0;
  }

int memql(cell obj, cell set)

  {
  while(set!=NIL)
    {
    if (eql(obj,car(set)))
      return 1;
    set = cdr(set);
    }
  return 0;
  }

int memqual(cell obj, cell set)

  {
  while(set!=NIL)
    {
    if (equal(obj,car(set)))
      return 1;
    set = cdr(set);
    }
  return 0;
  }


cell mqfind(cell what, cell where)

  {
  while(where!=NIL)
    {
    if(eq(what,car(where)))
      return where;
    where = cdr(where);
    }
  return NIL;
  }	
		 
cell mqlfind(cell what, cell where)

  {
  while(where!=NIL)
    {
    if(eql(what,car(where)))
      return where;
    where = cdr(where);
    }
  return NIL;
  }	

cell mqualfind(cell what, cell where)

  {
  while(where!=NIL)
    {
    if(equal(what,car(where)))
      return where;
    where = cdr(where);
    }
  return NIL;
  }	
		 
		 

cell insertq(cell obj, cell set)

  {
  if(memq(obj,set))
    return set;
  else
    return cons(obj,set);
  }

cell unionq(cell set1, cell set2)

  {
  while(set1!=NIL)
    {
    set2 = insertq(car(set1), set2);
    set1 = cdr(set1);
    }
  return set2;
  }

cell intersectq(cell set1, cell set2)

  {
  register cell rv;

  rv = NIL;
  while(set1!=NIL)
    {
    if (memq(car(set1), set2))
      rv = cons(car(set1), rv);
    set1 = cdr(set1);
    }
  return rv;
  }


/* special type of lists: FEL --- fuzzy elements lists
 * ( {symbol [n]}* ) where n is the fuzzy appartenence coeficient
 * and is a number between 1 and 100; it's absence means 100
 */

int is_partial(cell fel)

  /* returns 1 if appartenence of car(fel) to fel is in the list */

  {
  if(is_cons(fel) && is_cons(cdr(fel)) && is_binary(cadr(fel)))
    return 1;
  return 0;  /* i.e. it is 100% by default */
  }

int is_fel(cell fel)  /* returns 0 if it's not, 1 if it is
                                * and 2 if it is not minimal, i.e.
                                * one could optimize it by the use of
                                * fel_minimize
                                */

  {
  int rv = 1;

  while(fel!=NIL)
    {
    if(!is_symbol(car(fel)))
      return 0;
    if(is_partial(fel))
      {
      if(binary(cadr(fel))==100)
        rv = 2;
      fel = cdr(fel);
      }
    fel = cdr(fel);
    } 
  return rv;
  }

cell fel_drop(cell fel)  /* the list and the coefficients are freed */
                         /* to the cell memory allocator */

  {
  cell rv;

  rv = fel;
  while(rv!=NIL)
    {
    if(is_binary(car(rv)))
      ungcell(car(rv));
    fel = cdr(rv);
    ungcell(rv);
    rv = fel;
    }
  }

cell fel_minimize(cell fel)

  /* given a newly built fel it throws away 
   * those items containing an appartenence of 100%
   */

  {
  cell loc;
  cell rv;

  rv = fel;
  while(fel!=NIL)
    {
    if(is_partial(fel) && binary(cadr(fel))==100)
      {
      ungcell(cadr(fel));
      loc = cddr(fel);
      ungcell(cdr(fel));
      cdr(fel) = loc;
      }
    fel = cdr(fel);
    }
  return rv;
  }

cell fel_shrink(cell fel)  /* k x k y -> k x if x<y */
                           /* the top of the list and the apartenance
                            * coefficients are built anew
                            */

  {
  cell rv;

  rv = NIL;
  for(; fel!=NIL; fel = cdr(fel))
    {
    if(is_binary(car(fel)))
      continue;
    if(memqual(car(fel), rv))
      {
      if(is_partial(fel) && is_partial(rv))
        {
        if(binary(cadr(rv))<binary(cadr(fel)))
          binary(cadr(rv)) = binary(cadr(fel));
        continue;
        }		  
      if(is_partial(rv) && !is_partial(fel))
        binary(cadr(rv)) = 100;  
          /* this will be extruded by fel_minimize */
      continue;
      }
    else
      {
      if(is_partial(fel))
        rv = cons(make(BINARY, binary(cadr(fel))), rv);
  		rv = cons(car(fel), rv);
      }
    }  
  }

int fel_member(cell what, cell fel)  

   /* returns 0 if not present and 1--100 the degree of presence */

  {
  for(; fel!=NIL; fel = cdr(fel))
    {
    if(is_binary(car(fel)))
      continue;
    if(equal(what, car(fel)))
      {
      if(is_partial(fel))
        return (int)binary(cadr(fel));
      return 100;
      }
    }
  return 0;
  }


cell fel_insert(cell fwhat, cell fwhere)

  /* the first item of the fel fwhat is inserted in fwhere */

  {
  cell c;

  if((c = mqualfind(car(fwhat), fwhere))!=NIL)
	 {
    if(is_partial(fwhat) && is_partial(c))
  		{
      if(binary(cadr(fwhat)) > binary(cadr(c)))
        binary(cadr(c)) = binary(cadr(fwhat));
      return fwhere;
      }
    if(is_partial(c) && !is_partial(fwhat))
      {
      binary(cadr(c)) = 100;
      return fwhere;
      }
    return fwhere;
    }
  if(is_partial(fwhat))
    return cons(car(fwhat), cons(make(BINARY, binary(cadr(fwhat))), fwhere));
  else
    return cons(car(fwhat), fwhere);
  }

cell fel_jinsert(cell what, cell where, int coef)

  {
  cell c;

  if(coef==0)
    return where;
  if((c=mqualfind(what, where))!=NIL)
    {
    if(!is_partial(c))
      return where;
    if(coef>binary(cadr(c)))
      binary(cadr(c)) = coef;
    }
  if(coef<100)
    where = cons(make(BINARY, coef), where);
  where = cons(what, where);
  return where;
  }

cell fel_copy(cell fel)

  {
  if (fel==NIL)
    return NIL;
  if(is_binary(car(fel)))
    return cons(make(BINARY, binary(car(fel))), fel_copy(cdr(fel)));
  else
    return cons(car(fel), copy(cdr(fel)));
  }


cell fel_union(cell f1, cell f2)
  
  {
  f2 = fel_copy(f2);  
    /* THIS maked the result ungstructable
     * without loosing f2 
     */


  for(; f1!=NIL; f1 = cdr(f1))
    {
    if(is_binary(car(f1)))
      continue;
    f2 = fel_insert(f1, f2);
    }
  return f2;
  }

cell fel_intersect(cell f1, cell f2)

  {
  cell rv;
  int t, z;

  rv = NIL;
  for(; f1!=NIL; f1=cdr(f1))
    {
    if(is_binary(car(f1)))
      continue;
    if(t = fel_member(car(f1), f2))
      {
      z = fel_member(car(f1), f1);
      if(t>z) 
        t = z;
      if(t<100)
        rv = cons(make(BINARY, t), rv);
      rv = cons(car(f1), rv);
      }
    }
  return rv;
  } 

int fel_anytoins(cell l1, cell l2)

  /* 1 if any of l1's are to be inserted in l2 */

  {
  int t, z;

  for(; l1!=NIL; l1 = cdr(l1))
    {
    if(is_binary(car(l1)))
      continue;
	 t = fel_member(car(l1), l1);
    z = fel_member(car(l1), l2);
    if (t>z)
		return 1;
    }
  return 0;  /* none is so prezent */
  }

long fel_length(cell fel)

  {
  long rv;

  rv = 0;
  for(; fel!=NIL; fel = cdr(fel))
    if(!is_binary(car(fel)))
      rv ++;
  return rv;
  }

long fel_size(cell fel)

  {
  long rv;

  rv = 0;
  for(; fel!=NIL; fel = cdr(fel))
    if(!is_binary(car(fel)))
      rv += fel_member(car(fel), fel);
  return rv;
  }



/* End of the sets subsystem */


