/*    ILFIVE lisp objects library -- time objects
 *
 *    Copyright (c) 1991--1998 Alexandru Dan Corlan M.D.
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    In short, you can use and modify this program. You can
 *    distribute the program with or without your modifications if you want,
 *    but you must not require licence fees and must provide the
 *    full source code with the distribution. If you make an application program
 *    which just calls functions from this library you may distribute that
 *    application program without source code, but also as an linkable object file, 
 *    and must distribute also this library with source code.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>

#include <shpell5.h>

/* Static tables that give the names of days and months */

static char *romanian_days[7] =
  { "Duminica",
    "Luni",
    "Marti",
    "Miercuri",
    "Joi",
    "Vineri",
    "Simbata" };

static char *english_days[7] =
  { "Sunday",
    "Monday",
    "Thuesday",
    "Wednesday",
    "Thirsday",
    "Friday",
    "Saturday" };

static char *romanian_months[12] =
  { "Ianuarie",
    "Februarie",
    "Martie",
    "Aprilie",
    "Mai",
    "Iunie",
    "Iulie",
    "August",
    "Septembrie",
    "Octombrie",
    "Noiembrie",
    "Decembrie" };

static char *english_months[12] =
  { "January",
    "February",
    "March",
    "April",
    "May",
    "June",
    "July",
    "August",
    "September",
    "October",
    "November",
    "December" };

static int days_in_a_month[12] = {
  31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

static char **crt_dnames = english_days;
static char **crt_mnames = english_months;

clrtime(ISSTIME *t)

  {
  t->__year = NO_YEAR;
  t->__mmddhh = (NO_MONTH << 10) | (NO_DAY << 5) | NO_HOUR;
  t->__minsec = (NO_MIN << 6) | NO_SEC;
  t->__szms = (NO_ZONE << 10) | NO_MILISEC;
  }

static int _time_parse( /* returns 0 if OK, and various negative numbers if error */
  ISSTIME *t,           /* place to store value */
  char *s               /* string to parse */
  )

  {
  int signf, bcf, number1, number2;

  clrtime(t);

  signf = (*s == '-');
  bcf = (*s == '~');

  if (bcf && !isdigit(*(s+1)))
    return -1;  /* syntax error, B.C. specified but no year */
  if (bcf || signf)
    s++;
  set_sign(t, signf);
  if(isdigit(*s))
    {
    number1 = atoi(s);
    while(isdigit(*s))
      s++;
    switch(*s)
      {
      case '/':  /* number1 might be an year or a month */
        s++;
        if(isdigit(*s))
          {
          number2 = atoi(s);
          while(isdigit(*s))
            s++;
          switch(*s)
            {
            case '/':  /* number2 is certainly a month */
              if(bcf)
                set_year(t, -number1);
              else
                set_year(t, number1);
              if(!may_be_month(number2))
                return -100;  /* invalid month */
              set_month(t, number2);
              s++;
              goto addday;
            case ';': /* number2 is certainly a day */
              if(bcf) return -3;  /* year must be specified for time before Christ */
              if(!may_be_month(number1))
                return -100;
              set_month(t, number1);
              if(!may_be_day(number2))
                return -101;
              set_day(t, number2);
              s++;
              goto addhour;
            case '\0':
              if(bcf)  /* it must be year/month */
                {
                set_year(t, -number1);
              if(!may_be_month(number2))
                return -100;
                set_month(t, number2);
                return 0;
                }
              return -4;  /* cannot tell difference between year and month */
            default:
              return -2;
            }
          }
        /* if it was not a digit */
        if(bcf && !*s)
          {
          set_year(t, -number1);
          return 0;
          }
        return -2;
      case ';':  /* number1 is certainly a day */
        s++;
        if(!may_be_day(number1))
          return -101;
        set_day(t, number1);
        goto addhour;
      case ':':  /* number1 may be an hour or a minute */
        s++;
        if(isdigit(*s))
          {
          number2 = atoi(s);
          while(isdigit(*s))
            s++;
          switch(*s)
            {
            case ':':  /* number1 was an hour */
              if(!may_be_hour(number1))
                return -102;
              set_hour(t, number1);
              if(!may_be_min(number2))
                return -103;
              set_min(t, number2);
              s++;
              goto addsec;
            case '.': /* number1 was a minute */
              if(!may_be_min(number1))
                return -103;
              set_min(t, number1);
              if(!may_be_sec(number2))
                return -104;
              set_sec(t, number2);
              s++;
              goto addmilisec;
            default:
              return -2;
            }
          }
        return -2;
      case '.':  /* number1 is a seccond */
        if(!may_be_sec(number1))
          return -104;
        set_sec(t, number1);
        s++;
        goto addmilisec;
      case '\0': /* CONVENTION: consider it an year */
        if (bcf)
          set_year(t, -number1);
        else
          set_year(t, number1);
        return 0;  /* OK, it was just an year */
      default:
        return -2;  /* invalid character in input */
      }
    }
  /* string does not start with a digit */
  switch(*s)
    {
    case ';':
      s++;
      goto addhour;
    case '.':
      s++;
      goto addmilisec;
    case '[':
      s++;
      if(signf || bcf)
        return -6;   /* zone can appear alone only without B.C. or minus */
      goto addzone;  /* permitted possibility to only specify one zone */
    case '/':
      s++;
      if(!isdigit(*s))
        return -5;  /* digit expected */
      number1 = atoi(s);
      while(isdigit(*s))
        s++;
      switch(*s)
        {
        case '/':  /* it was a month */
          if(!may_be_month(number1))
            return -100;
          set_month(t, number1);
          s++;
          goto addday;
        case ';':  /* it was a day */
          if(!may_be_day(number1))
            return -101;
          set_day(t, number1);
          s++;
          goto addhour;
        default:
          return -2;
        }
    case ':':
      s++;
      if(!isdigit(*s))
        return -5;  /* digit expected */
      number1 = atoi(s);
      while(isdigit(*s))
        s++;
      switch(*s)
        {
        case ':': /* it was a minute */
          if(!may_be_min(number1))
            return -103;
          set_min(t, number1);
          s++;
          goto addsec;
        case '.':
          if(!may_be_sec(number1))
            return -104;
          set_sec(t, number1);
          s++;
          goto addmilisec;
        default:
          return -2;
        }
    default:
      return -2;
    }
  return -2;

addday:
  if(!*s) return 0;
  if(!isdigit(*s)) return -2;
  set_day(t, number1 = atoi(s));
  if(!may_be_day(number1))
    return -101;
  while(isdigit(*s))
    s++;
  if(!*s) return 0;
  if(*s!=';') return -2;
  s++;
addhour:
  if(!*s) return 0;
  if(!isdigit(*s)) return -2;
  set_hour(t, number1 = atoi(s));
  if(!may_be_hour(number1))
    return -102;
  while(isdigit(*s))
    s++;
  if(!*s) return 0;
  if(*s!=':') return -2;
  s++;
/*addmin: */
  if(!*s) return 0;
  if(!isdigit(*s)) return -2;
  set_min(t, number1 = atoi(s));
  if(!may_be_min(number1))
    return -103;
  while(isdigit(*s))
    s++;
  if(!*s) return 0;
  if(*s!=':') return -2;
  s++;
addsec:
  if(!*s) return 0;
  if(!isdigit(*s)) return -2;
  set_sec(t, number1 = atoi(s));
  if(!may_be_sec(number1))
    return -104;
  while(isdigit(*s))
    s++;
  if(!*s) return 0;
  if(*s!='.') return -2;
  s++;
addmilisec:
  if(!*s) return 0;
  if(!isdigit(*s)) return -2;
  set_milisec(t, number1 = atoi(s));
  if(!may_be_milisec(number1))
    return -105;
  while(isdigit(*s))
    s++;
  if(!*s) return 0;
  if(*s!='[') return -2;
  s++;
  if(!has_year(t))
    return -7;  /* only moments and zones can have a zone specified */
addzone:
  if(!isdigit(*s)) return -2;
  set_zone(t, number1 = atoi(s));
  if(!may_be_zone(number1))
    return -106;
  while(isdigit(*s))
    s++;
  if(*s!=']') return -2;
  if(*s && !isspace(*s))
    return -8;  /* invalid end of time expression */
  return 0;
  }

int possible_days(ISSTIME *t)

  {
  if(!has_year(t))
    return 30;
  if(month(t)!=1)
    return days_in_a_month[month(t)];
  if(is_bisect(t))
    return 29;
  return 28;
  }

int checktime(ISSTIME *t)

  {
  if (!has_year(t))
    return 0;  /* it is a duration, of no importance */
  if(month(t)!=1)
    {
    if(days_in_a_month[month(t)] > day(t))
      return 0;
    return 1;
    }
  if(is_bisect(t))
    return day(t) < 29;
  else
    return day(t) < 28;

/* the Gregorian days and also some special secconds might be excluded here */

  }

int time_parse(ISSTIME *t, char *s)

  {
  int r;

  r = _time_parse(t,s);
  if (r<0)
    return r;    /* failed from then */
  return checktime(t);  /* a negative result = no such date */
  }


/* simple local_itoa */

local_itoa(int n, char *rev, int sz)

{
char bf[20];

sprintf(bf, "%%%dd", sz);
sprintf(rev, bf, n);
}

/* the following function returns a string containing a rereadable
 * description of the given time, or NULL if not possible
 */



char *time_unparse(ISSTIME *t)

  {
  static char rv[32];
  char *rev;
  int p, startf;

  rev = rv;
  startf = 0;
  if(has_sign(t))
    *rev++ = '-';
  if(has_year(t))
    {
    startf = 1;
    if((p = year(t))<0)
      *rev++ = '~';
    local_itoa(abs(p), rev, 10);
    while(*rev)
      rev++;
    }
  if(has_month(t))
    {
    startf = 1;
    *rev++ = '/';
    local_itoa(month(t), rev, 10);
    while(*rev)
      rev ++;
    }
  else
    if(startf)
      {
      *rev++ = '/';
      *rev = 0;
      return rv;
      }
  if(has_day(t))
    {
    startf = 1;
    *rev++ = '/';
    local_itoa(day(t), rev, 10);
    while(*rev)
      rev++;
    }
  else
    if(startf)
      {
      *rev++ = '/';
      *rev = 0;
      return rv;
      }
  if(has_hour(t))
    {
    startf = 1;
    *rev++ = ';';
    local_itoa(hour(t), rev, 10);
    while(*rev)
      rev++;
    }
  else
    if(startf)
      {
      *rev++ = ';';
      *rev = 0;
      return rv;
      }
  if(has_min(t))
    {
    startf = 1;
    *rev++ = ':';
    local_itoa(min(t), rev, 10);
    while(*rev)
      rev++;
    }
  else
    if(startf)
      {
      *rev++ = ':';
      *rev = 0;
      return rv;
      }
  if(has_sec(t))
    {
    startf = 1;
    *rev++ = ':';
    local_itoa(sec(t), rev, 10);
    while(*rev)
      rev++;
    }
  else
    if(startf)
      {
      *rev++ = ':';
      *rev = 0;
      return rv;
      }
  if(has_milisec(t))
    {
    startf = 1;
    *rev++ = '.';
    if(milisec(t) < 100)
      *rev++ = '0';
    if(milisec(t) < 10)
      *rev++ = '0';
    local_itoa(milisec(t), rev, 10);
    while(*rev)
      rev++;
    }
  else
    if(startf)
      {
      *rev++ = '.';
      *rev = 0;
      return rv;
      }
  if(has_zone(t))
    {
    *rev++ = '[';
    local_itoa(zone(t), rev, 10);
    while(*rev)
      rev++;
    *rev++ = ']';
    *rev = 0;
    return rv;
    }
  *rev = 0;
  return rv;  /* no component in this time struct */
  }

/* the following function sets the spoken language of the user */

int set_language(int lang)

  {  /* returns 0 if OK, -1 if error */
  if(lang==ENGLISH)
    {
    crt_dnames = english_days;
    crt_mnames = english_months;
    return 0;
    }
  if(lang==ROMANIAN)
    {
    crt_dnames = romanian_days;
    crt_mnames = romanian_months;
    return 0;
    }
  return -1;
  }

/* function to return the day of the week [0 for Sunday, 6 for Saturday] */

int week_day(ISSTIME *t)

  {
  return 0;  /* not written yet */
  }

/* utility functions for time_sprintf */

char *catstr(char *dest, char *arg, int size)

  {
  int i;
  int n;

  n = strlen(arg);
  if(n>size)
    {
    n = size;
    arg[size] = 0;
    }
  for(i=size; i<n; i++)
    *dest++ = ' ';
  strcpy(dest, arg);
  dest += strlen(dest);
  return dest;
  }

char *catpad(char *dest, int sz, int ch)

  {
  int i;

  for(i=0; i<sz; i++)
    *dest++ = ch;
  *dest = 0;
  return dest;
  }

char *catint(char *dest, int val, int sz, int zero)

  {
  int i, n;
  static char b[20];

  local_itoa(val, b, 10);
  n = strlen(b);
  if (n>sz)
    n = sz;
  for(i=n; i<sz; i++)
    if(zero)
      *dest++ = '0';
    else
      *dest++ = ' ';
  strcpy(dest, b + strlen(b) - n);
  dest += strlen(dest);
  return dest;
  }

/* The time formatting function */

char *time_sprintf(char *fm, ISSTIME *t)

  {
  static char rv[128];  /* maximum size of a full time description */
  char *rev;
  int zerof;
  int sizef;
  int n, i;

  rev = rv;
  while(*fm && *fm != '~')
    {
    if(!isdigit(*fm) && !isalpha(*fm))
      {
      *rev++ = *fm++;
      continue;
      }
    zerof = sizef = 0;
    if(isdigit(*fm))
      {
      while(*fm=='0')
        {
        zerof = 1;
        fm ++;
        }
      if(isdigit(*fm))
        sizef = atoi(fm);
      while (isdigit(*fm))
        fm ++;
      }
    if(!isalpha(*fm))
      return NULL; /* error in format string */
    switch(fm[0])
      {
      case 'g':  /* sign */
        if(sizef)
          for(i=0; i<sizef-1; i++)
            *rev++ = ' ';
        if(has_sign(t))
          *rev++ = '-';
        else
          if(sizef)
            *rev++ = ' ';
        fm++;
        break;
      case 'y':  /* year */
        if(has_year(t))
          rev = catint(rev, abs(year(t)), sizef, zerof);
        else
          rev = catpad(rev, sizef, zerof ? '0' : ' ');
        fm ++;
        break;
      case 'm':  /* month */
        if(has_month(t))
          rev = catint(rev, 1+month(t), sizef, zerof);
        else
          rev = catpad(rev, sizef, zerof ? '0' : ' ');
        fm++;
        break;
      case 'M':  /* textual month */
        if(has_month(t))
          rev = catstr(rev, crt_mnames[month(t)], sizef);
        else
          rev = catpad(rev, sizef, ' ');
        fm++;
        break;
      case 'w':
        if((n = week_day(t)) < 0) return NULL;
        rev = catint(rev, n, sizef, zerof);
        fm++;
        break;
      case 'W':
        if((n = week_day(t)) < 0) return NULL;
        rev = catstr(rev, crt_dnames[n], sizef);
        fm++;
        break;
      case 'd':
        if(has_day(t))
          rev = catint(rev, day(t), sizef, zerof);
        else
          rev = catpad(rev, sizef, zerof ? '0' : ' ');
        fm++;
        break;
      case 'h':
        if(has_hour(t))
          rev = catint(rev, hour(t), sizef, zerof);
        else
          rev = catpad(rev, sizef, zerof ? '0' : ' ');
        fm++;
        break;
      case 't':
        if(has_min(t))
          rev = catint(rev, min(t), sizef, zerof);
        else
          rev = catpad(rev, sizef, zerof ? '0' : ' ');
        fm ++;
        break;
      case 's':
        if(has_sec(t))
          rev = catint(rev, sec(t), sizef, zerof);
        else
          rev = catpad(rev, sizef, zerof ? '0' : ' ');
        fm ++;
        break;
      case 'l':
        if(has_milisec(t))
          rev = catint(rev, milisec(t), sizef, zerof);
        else
          rev = catpad(rev, sizef, zerof ? '0' : ' ');
        fm ++;
        break;
      case 'z':
        if(has_zone(t))
          rev = catint(rev, zone(t), sizef, zerof);
        else
          rev = catpad(rev, sizef, zerof ? '0' : ' ');
        fm ++;
        break;
      case 'b':
      case 'c':
        if(has_year(t) && year(t)<0)
          rev = catstr(rev, "B.C.", sizef);
        else
          rev = catpad(rev, sizef, ' ');
        fm ++;
        break;
      default:
        return NULL;
      }
    }
  *rev = 0;
  return rv;
  }

/* The following function, which reads system time
 * only works with TopSpeed C.
 */

int set_to_now(ISSTIME *t)

  {
  static struct tm tt;
  time_t tmt;

  tmt = time(NULL);
  tt = * localtime(&tmt);
  clrtime(t);
  set_year(t,tt.tm_year+1900);
  set_month(t,tt.tm_mon);
  set_day(t,tt.tm_mday);
  set_hour(t,tt.tm_hour);
  set_min(t,tt.tm_min);
  set_sec(t,tt.tm_sec);
  return 0;
  }


